# -*- coding: utf-8 -*-
"""
@File    : extractors.py
@Date    : 2024/6/29 下午1:57
@Author  : yintian
@Desc    :  Paser mode by bricks
"""


import collections
import inspect
import itertools
import math
import re
import time
from typing import Any, Optional, Callable, Union, Type, Literal, List

from ytools.utils import magic
from ytools.utils.magic import require

require('jmespath>=1.0.1')
require('jsonpath>=0.82.2')
require('lxml>=4.9.3')

import jmespath  # noqa
import jsonpath  # noqa
from lxml import etree  # noqa


class Rule:

    def __init__(
            self,
            exprs: Optional[Union[str, Callable]] = "",
            condition: Optional[Callable] = None,
            pre_script: Optional[Callable] = None,
            post_script: Optional[Callable] = None,
            is_array: Optional[bool] = None,
            acquire: bool = True,
            options: Optional[dict] = None,
            const: Optional[Any] = ...,
            default: Optional[Any] = None,
            engine: Optional[Union[Type["Extractor"], str]] = None,
    ):
        self.exprs = exprs
        self.condition = condition
        self.pre_script = pre_script
        self.post_script = post_script
        self.is_array = is_array
        self.allow_nulls = acquire
        self.options = options or {}
        self.const = const
        self.default = default
        self.engine: Optional["Extractor"] = None
        self.build_engine(engine)

    @classmethod
    def parse(cls, value, engine=None):
        if isinstance(value, (cls, Group)):
            value.build_engine(engine)
            return value
        else:
            return cls(exprs=value, engine=engine)

    def build_engine(self, value):
        if self.engine or not value:
            return

        if isinstance(value, Extractor) or (inspect.isclass(value) and issubclass(value, Extractor)):
            self.engine = value

        elif isinstance(value, str):
            self.engine = globals()[f'{value.title()}Extractor']

        else:
            raise ValueError("engine must be a string or Extractor instance")

    def apply(self, obj, strict: Literal["ignore", "raise"] = "ignore"):
        # 常量
        if self.const is not ...:
            return self.const

        # 前置处理脚本
        if callable(self.pre_script):
            obj = self.pre_script(obj)

        # 匹配结果
        res = self.engine.extract(obj, self.exprs, **self.options)

        # 设置默认值
        if res in self.engine.empty:
            if strict == "raise":
                raise Extractor.Empty
            res = self.default

        # 后置处理脚本
        if callable(self.post_script):
            res = self.post_script(res)

        # 引擎适配逻辑
        if self.is_array is False:
            res = magic.first(res)

        elif self.is_array is True:
            res = magic.iterable(res)

        elif self.engine.is_array and isinstance(res, list) and len(res) == 1:
            res = res[0]

        return res

    def is_work(self, obj):
        if self.condition is None:
            return True

        assert callable(self.condition), "condition must be a callable object"
        return bool(self.condition(obj))

    def __str__(self):
        return f"<Rule {self.exprs}>"

    def __or__(self, other):
        return Group([self, other])


class Group:
    """
    用多多个 rule 的 or 结果

    """

    def __init__(
            self,
            exprs: List[Union[Rule, 'Group']],
            pre_script: Optional[Callable] = None,
            post_script: Optional[Callable] = None,
    ):
        self.exprs = magic.iterable(exprs)
        self.pre_script = pre_script
        self.post_script = post_script

    def apply(self, obj, strict: Literal["ignore", "raise"] = "ignore"):
        # 前置处理脚本
        if callable(self.pre_script):
            obj = self.pre_script(obj)

        for exprs in self.exprs:
            try:
                res = exprs.apply(obj, strict="raise")
                break
            except Extractor.Empty:
                pass
        else:
            if strict == 'ignore':
                res = self.exprs[-1].default
            else:
                raise Extractor.Empty

        # 后置处理脚本
        if callable(self.post_script):
            res = self.post_script(res)

        return res

    def is_work(self, obj):
        for exprs in self.exprs:
            if exprs.is_work(obj):
                return True
        else:
            return False

    def build_engine(self, engine):
        for exprs in self.exprs:
            exprs.build_engine(engine)

    def __or__(self, other):
        return Group([self, other])


class Extractor:
    """
    提取器抽象类

    """

    is_array = False
    empty = (None,)

    class Empty(Exception):
        ...

    @classmethod
    def extract(cls, obj: Any, exprs: str, **kwargs):
        raise NotImplementedError

    @staticmethod
    def fmt(obj, **kwargs):
        raise NotImplementedError

    @classmethod
    def extract_first(cls, obj: Any, exprs: str, default: Any = None, **kwargs):
        res = cls.extract(
            obj=obj,
            exprs=exprs,
            **kwargs
        )
        return magic.first(res, default=default)

    @classmethod
    def match(cls, obj: Any, rules: Union[dict, list]):
        rows = []

        def tree():
            return collections.defaultdict(tree)

        for rule in magic.iterable(rules):
            vessel = tree()

            for item in cls._match(
                    obj=[obj],
                    rules=rule,
            ):
                key, value, rkey1, rkey2, rkey3 = item
                if value is not None: vessel[rkey1][rkey2][rkey3][key] = value

            cleans = []
            for k1, v1 in vessel.items():
                skeys = sorted(v1.keys())
                for index, skey in enumerate(skeys[:-1]):
                    d2 = v1[skeys[index + 1]]
                    for k2, v2 in v1.pop(skey).items():
                        hit = False
                        for k3, v3 in d2.items():
                            if k3.startswith(k2):
                                hit = True
                                v3.update({**v2, **v3})

                        if not hit: d2.update({k2: v2})

                else:
                    # 不同层级执行笛卡尔积
                    cleans = [
                        {**i[1], **i[0]}
                        for i in itertools.product(v1[skeys[-1]].values(), cleans or [{}])
                    ]

            if isinstance(rules, list):
                rows.extend(cleans)
            else:
                rows = cleans

        return rows

    @classmethod
    def _match(cls, obj: Any, rules: dict, rkey="", pkey="", deep=0):
        if obj in [None, []]: obj = [None]
        obj = magic.iterable(obj, enforce=object, exclude=(set, list, tuple))

        def get_rkey(*keys, splice="."):
            return splice.join(filter(bool, keys))

        number_of_digits = math.ceil(math.log(len(obj), 10))

        for index, _obj in enumerate(obj):

            for k, v in rules.items():

                if not isinstance(v, dict):
                    rule = Rule.parse(v, engine=cls)
                    _rkey = (
                        rkey.split(".")[0],
                        pkey.count(".") + deep,
                        get_rkey(pkey, str(index).zfill(number_of_digits)) if deep else pkey
                    )

                    if rule.exprs == '@index':
                        yield k, index, *_rkey

                    elif rule.exprs == '@ts':
                        yield k, time.time(), *_rkey

                    elif rule.exprs == '@date':
                        yield k, time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()), *_rkey

                    elif rule.is_work(_obj):
                        _unpack_obj = rule.apply(_obj)

                        if k == '@unpack':

                            if isinstance(_unpack_obj, dict):
                                for _k, _v in _unpack_obj.items():
                                    yield _k, _v, *_rkey

                            elif isinstance(_unpack_obj, (list, tuple, set)):

                                for _k, _v in enumerate(_unpack_obj):
                                    yield _k, _v, *_rkey

                            else:
                                for _k, _v in [(k, _unpack_obj)]:
                                    yield _k, _v, *_rkey

                        else:
                            yield k, _unpack_obj, *_rkey

                # 代表进入下一级别 / 提取功能参数
                else:

                    rule = Rule.parse(k, engine=cls)
                    if not rule.is_work(_obj):
                        sub_obj = rule.apply(_obj)

                        for i in cls._match(
                                obj=sub_obj,
                                rules=v,
                                rkey=get_rkey(rkey, rule.exprs),
                                pkey=get_rkey(pkey, str(index).zfill(number_of_digits)),
                                deep=bool(isinstance(sub_obj, list))
                        ):
                            yield i


class XpathExtractor(Extractor):
    """
    xpath 提取器

    """
    is_array = True
    empty = (None, [])

    @classmethod
    def extract(cls, obj: Union[etree.HTML, str], exprs: str, parser=None, base_url=None):
        obj = cls.fmt(obj, parser=parser, base_url=base_url)
        ret = obj.xpath(exprs, parser=parser, base_url=base_url)
        return [i.strip() if isinstance(i, str) else i for i in ret]

    @classmethod
    def fmt(cls, obj, parser=None, base_url=None):

        if isinstance(obj, str):
            if obj.strip().startswith("<?xml "):
                obj = etree.fromstring(obj, parser=parser, base_url=base_url)
            else:
                obj = etree.HTML(obj.encode(), parser=parser, base_url=base_url)
        return obj


class JsonExtractor(Extractor):
    """
    json jmespath 提取器

    """
    is_array = False
    empty = (None,)

    @classmethod
    def extract(cls, obj: Union[dict, list, str], exprs: str, jsonp=False, errors='strict', options=None):
        obj = cls.fmt(obj, jsonp=jsonp, errors=errors)
        ret = jmespath.search(
            data=obj,
            expression=exprs,
            options=options,
        )
        return ret

    @staticmethod
    def fmt(obj, **kwargs):
        if isinstance(obj, str):
            obj = magic.json_or_eval(
                text=obj,
                **kwargs
            )

        return obj


class JsonpathExtractor(Extractor):
    is_array = False
    empty = (None,)

    @classmethod
    def extract(cls, obj: Union[dict, list, str], exprs: str, jsonp=False, errors='strict', result_type='VALUE',
                debug=0, use_eval=True):
        obj = cls.fmt(obj, jsonp=jsonp, errors=errors)
        ret = jsonpath.jsonpath(
            obj=obj,
            expr=exprs,
            result_type=result_type,
            debug=debug,
            use_eval=use_eval
        )
        if ret is False: ret = None
        return ret

    @staticmethod
    def fmt(obj, **kwargs):
        if isinstance(obj, str):
            obj = magic.json_or_eval(
                text=obj,
                **kwargs
            )

        return obj


class RegexExtractor(Extractor):
    """
    正则提取器

    """
    is_array = True
    empty = (None, [], "")

    @classmethod
    def extract(cls, obj: str, exprs: str, flags=0):
        obj = cls.fmt(obj)
        return re.findall(exprs, obj, flags=flags)

    @staticmethod
    def fmt(obj, **kwargs):  # noqa
        if not isinstance(obj, str):
            if isinstance(obj, etree._Element):  # noqa
                obj = etree.tostring(obj).decode()
            else:
                obj = str(obj) if obj is not None else ""

        return obj


if __name__ == '__main__':
    data = {
        "name1": "name1",
        "name2": "name2",
        "name3": "name3",
        # "name4": "name4",
    }

    print(JsonExtractor.match(data, {"res": Rule("name4") | Rule("name3") | Rule("name2") | Rule("name1")}))



# -*- coding: utf-8 -*-
"""
@File    : route.py
@Date    : 2024/8/12 上午9:17
@Author  : yintian
@Desc    : 
"""
from ytools.utils.magic import require

require('DrissionPage>=4.0.5.6')

import base64
import json
import re
import time
from collections import defaultdict
from hashlib import md5

from DrissionPage._base.driver import Driver  # noqa
from DrissionPage._pages.chromium_page import ChromiumPage  # noqa

from ytools.utils.magic import result


class Full:

    def __init__(self, **kwargs):
        self.options = kwargs.copy()
        self.options.pop('driver', None)
        request: dict = kwargs.get('request')
        self.network_id = kwargs.get('networkId')
        self.driver: Driver = kwargs.get('driver')
        self.frame_id = kwargs.get('frameId')
        self.request_id = kwargs.get('requestId')
        self.url = request.get('url')
        self.method = request.get('method')
        self.headers = request.get('headers')
        self.post_data = request.get('postData')
        self.extra = RouteRequestExtra(**kwargs)
        self.full_content: bytes = b''
        self.full_headers: dict = {}
        self.full_error: str = ''
        self.full_status_code: int = 0
        self.fulls = {
            # interceptResponse: bool, full request use
            # responsePhrase: str, full response use
        }
        self.raw = self.fingerprint

    @property
    def fingerprint(self):
        return md5(
            json.dumps(
                {
                    'requestId': self.request_id,
                    'url': self.url,
                    'method': self.method,
                    'headers': self.headers,
                    'post_data': self.post_data
                }
            ).encode()
        ).hexdigest()

    @property
    def flag(self):
        return any(
            [
                self.raw != self.fingerprint,
                self.full_status_code,
                self.full_content,
                self.full_headers,
                self.full_error,
                self.fulls,
            ]
        )

    @property
    def data(self):
        return {
            "requestId": self.request_id,
            **self.fulls
        }

    @property
    def continue_data(self):
        return {
            **self.data,
            "url": self.url,
            "method": self.method,
            "postData": self.post_data or '',
            "headers": [{'name': k, 'value': v} for k, v in self.headers.items()],
        }

    @property
    def full_header_list(self):
        header_list = []
        for k, v in self.full_headers.items():
            header_list.append({'name': k, 'value': v})
        return header_list

    @property
    def full_data(self):
        data = {
            **self.data,
            "responseCode": self.full_status_code or 200
        }
        if self.full_headers:
            data['responseHeaders'] = self.full_header_list
        if self.full_content:
            data['body'] = base64.b64encode(self.full_content).decode()
        return data

    @property
    def error_data(self):
        if self.full_error not in ['Failed', 'Aborted', 'TimedOut', 'AccessDenied', 'ConnectionClosed',
                                   'ConnectionReset', 'ConnectionRefused', 'ConnectionAborted', 'ConnectionFailed',
                                   'NameNotResolved', 'InternetDisconnected', 'AddressUnreachable',
                                   'BlockedByClient', 'BlockedByResponse']:
            self.full_error = 'Aborted'
        data = {
            **self.data,
            "errorReason": self.full_error,
        }
        return data

    @property
    def full_response_data(self):
        return {
            **self.data,
            'responseHeaders': self.full_header_list
        }


class RouteRequestExtra:
    def __init__(self, **kwargs):
        self.resource_type = kwargs.get('resourceType')
        self.resource_error_reason = kwargs.get('responseErrorReason')
        self.response_status_code = kwargs.get('responseStatusCode')
        self.response_status_text = kwargs.get('responseStatusText')
        self.response_headers = kwargs.get('responseHeaders')
        self.redirect_request_id = kwargs.get('redirectedRequestId')
        self.redirect_url = kwargs.get('redirectUrl')
        self.is_download = kwargs.get('isDownload')
        self.is_navigation_request = kwargs.get('isNavigationRequest')
        self.auth_challenge = kwargs.get('authChallenge')
        self.is_download = kwargs.get('isDownload')
        request = kwargs.get('request')
        self.has_post_data = request.get('hasPostData')
        self.post_data_entries = request.get('postDataEntries')
        self.mixed_content_type = request.get('mixedContentType')
        self.initial_priority = request.get('initialPriority')
        self.referrer_policy = request.get('referrerPolicy')
        self.is_link_preload = request.get('isLinkPreload')
        self.trust_token_params = request.get('trustTokenParams')
        self.is_same_site = request.get('isSameSite')
        self.url_fragment = request.get('urlFragment')


class RouteRequest(Full):
    def continue_request(self):
        _method = 'Fetch.continueRequest'
        if not self.flag:
            continue_data = self.data
        elif self.full_error:
            _method = 'Fetch.failRequest'
            continue_data = self.error_data
        elif self.full_content or self.full_headers:
            _method = 'Fetch.fulfillRequest'
            continue_data = self.full_data
        else:
            continue_data = self.continue_data
        self.driver.run(
            _method,
            **continue_data
        )


class RouteResponse(Full):
    content: bytes = b''
    encoding = 'utf-8'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.request: RouteRequest = kwargs.get('_request')
        self.headers = {_['name']: _['value'] for _ in self.extra.response_headers}
        self.get_response_body()

    @property
    def text(self):
        return self.content.decode(encoding=self.encoding) if not self.full_content else self.full_content.decode(encoding=self.encoding)

    @text.setter
    def text(self, value):
        if isinstance(value, str):
            self.content = value.encode(encoding=self.encoding)
        elif isinstance(value, bytes):
            self.content = value
        else:
            try:
                self.content = json.dumps(value).encode(encoding=self.encoding)
            except Exception as e:
                raise e
        self.full_content = self.content

    def json(self):
        return json.loads(self.text)

    def get_response_body(self):
        if self.content:
            return self.content
        res = self.driver.run(
            'Fetch.getResponseBody',
            requestId=self.request_id
        )
        if body := res.get('body'):
            if res.get('base64Encoded'):
                self.content = base64.b64decode(body)
            else:
                if isinstance(body, str):
                    self.content = body.encode()
                elif isinstance(body, bytes):
                    self.content = body
                else:
                    raise ValueError(f'body 类型错误,应为 str, bytes, 获取到 {type(body)}')
        return self.content

    def continue_request(self, ):
        _method = 'Fetch.continueRequest'
        if not self.flag:
            continue_data = self.data
        elif self.full_error:
            _method = 'Fetch.failRequest'
            continue_data = self.error_data
        elif self.full_content:
            _method = 'Fetch.fulfillRequest'
            continue_data = self.full_data
        elif self.full_headers:
            _method = 'Fetch.continueResponse'
            continue_data = self.full_data
        else:
            continue_data = self.data
        self.driver.run(
            _method,
            **continue_data
        )


class Route:
    _patters = defaultdict(dict)
    _request = defaultdict(dict)
    _driver: list[Driver] = []

    @classmethod
    def start_by_driver(cls, driver: Driver):
        cls._driver.append(driver)
        cls.update_listen()

    @classmethod
    def update_listen(cls):
        patterns = []
        for patters, actions in cls._patters.items():
            if actions.get('on_request'):
                patterns.append({
                    "urlPattern": f"*{patters}*",
                    "requestStage": "Request"
                })
            if actions.get('on_response'):
                patterns.append({
                    "urlPattern": f"*{patters}*",
                    "requestStage": "Response"
                })
        for driver in cls._driver:
            driver.run(
                'Fetch.enable',
                patterns=patterns
            )
            driver.set_callback(
                'Fetch.requestPaused',
                cls.mock_func(
                    func=cls.request_pause,
                    driver=driver
                )
                # cls.request_pause
            )

    @staticmethod
    def mock_func(func, *args, **kwargs):
        def inner(**k):
            _kwargs = kwargs.copy()
            _kwargs.update(k or {})
            return result(
                func=func,
                args=args,
                kwargs=_kwargs,
            )

        return inner

    @classmethod
    def set(cls, action, target, func):
        if action not in cls._patters[target]:
            cls._patters[target][action] = []
        cls._patters[target][action].append(func)

    @classmethod
    def get(cls, action, target):
        if action not in cls._patters[target]:
            cls._patters[target][action] = []
        return cls._patters[target][action]

    @classmethod
    def pop(cls, action, target, func):
        if action not in cls._patters[target]:
            cls._patters[target][action] = []
        cls._patters[target][action].remove(func)

    @classmethod
    def on(cls, action, target, func):
        cls.set(f'on_{action}', target, func)
        cls.update_listen()

    @classmethod
    def format_request(cls, **kwargs):
        driver: Driver = kwargs.get('driver')
        request = RouteRequest(**kwargs)
        cls._request[request.request_id]['request'] = request
        cls._request[request.request_id]['driver'] = driver
        return request

    @classmethod
    def format_response(cls, **kwargs):
        response = RouteResponse(**kwargs)
        cls._request[response.request.request_id]['response'] = response
        return response

    @classmethod
    def request_pause(cls, **kwargs):
        request = cls.format_request(**kwargs)
        if request.extra.response_status_code:
            kwargs['_request'] = request
            response = cls.format_response(**kwargs)
            cls._on_response(request, response)
        else:
            cls._on_request(request)

    @classmethod
    def _on_request(cls, request: RouteRequest):
        for p, v in cls._patters.items():
            if re.search(p, request.url):
                for func in v.get('on_request', []):
                    cls.mock_func(
                        func=func,
                        request=request
                    )()
                    if request.flag:
                        break
        request.continue_request()

    @classmethod
    def _on_response(cls, request: RouteRequest, response: RouteResponse):
        for p, v in cls._patters.items():
            if re.search(p, response.url):
                for func in v.get('on_response', []):
                    cls.mock_func(
                        func=func,
                        request=request,
                        response=response
                    )()
                    if not response.flag:
                        break
        response.continue_request()


def mock_raw(request: RouteRequest):
    request.full_content = b""


def mock_res(response: RouteResponse):
    print()
    response.text = open('2.25.1.js').read().replace('2.25.1', '2.25.47') + "\n // 2.25.1"
    response.full_headers = {
        **response.headers,
        "content-length": str(response.full_content.__len__()),
    }
    print(123)


def test():
    from DrissionPage._configs.chromium_options import ChromiumOptions  # noqa
    opt = ChromiumOptions()
    opt.set_argument(' --disable-web-security', "")
    opt.set_browser_path("a")
    browser = ChromiumPage(opt)
    Route.start_by_driver(driver=browser.driver)
    # Route.on('request', '2.25.', mock_raw)
    Route.on('response', '2.25.', mock_res)
    browser.get(f'https://shopee.co.th/buyer/login')
    while True:
        time.sleep(1)


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
"""
@File    : magic_for_class.py
@Author  : yintian
@Date    : 2025/5/26 11:03
@Software: PyCharm
@Desc    : 
"""
import functools
import inspect
import json


class CustomMeta(type):
    _instances = {}

    def __new__(mcs, name, bases, attrs, **kwargs):
        __cls_kwargs = {'singleton': False, 'wrappers': '_when_', 'fix': {}}
        keys = ['singleton', 'wrappers', 'fix']
        # 获取单例参数和包装前缀
        for key in keys:
            if value := kwargs.get(key):
                __cls_kwargs[key] = value
                continue
            for base in bases:
                cls_kw = hasattr(base, '__cls_kwargs') and (getattr(base, '__cls_kwargs', {}))
                if cls_kw and (value := cls_kw.get(key)) is not None:
                    __cls_kwargs[key] = value
                    break
        attrs['__cls_kwargs'] = __cls_kwargs
        # 不修改方法，仅创建类
        cls = super().__new__(mcs, name, bases, attrs)
        return cls

    def __call__(cls, *args, **kwargs):
        # 获取单例参数和包装前缀
        cls_kwargs = getattr(cls, '__cls_kwargs', {})
        singleton = cls_kwargs.get('singleton', False)
        wrappers = cls_kwargs.get('wrappers', '_when_')
        fix = cls_kwargs.get('fix', {})

        def when_match(ins, attr_name):
            attr_value = getattr(ins, attr_name)
            if not callable(attr_value):
                return False
            wrapper_name = f'{wrappers}{attr_name}'
            wrapper_func = getattr(ins, wrapper_name, None)
            if wrapper_func and callable(wrapper_func):
                return True
            return False

        def when_fix(ins, attr_name):
            attr_value = getattr(ins, attr_name)
            wrapper_name = f'{wrappers}{attr_name}'
            wrapper_func = getattr(ins, wrapper_name)
            setattr(ins, attr_name, wrapper_func(attr_value))

        fix['when'] = {
            "match": when_match,
            "fix": when_fix
        }

        def get_key():
            if isinstance(singleton, str):
                if singleton == 'all':
                    # 使用所有参数生成键
                    init_signature = inspect.signature(cls.__init__)
                    bound_args = init_signature.bind(None, *args, **kwargs)
                    bound_args.apply_defaults()
                    param_dict = {k: v for k, v in bound_args.arguments.items() if k != 'self'}
                    # 使用 json.dumps 序列化，sort_keys 保证顺序一致，default=str 处理不可序列化对象
                    return json.dumps({'class': cls.__name__, 'params': param_dict}, sort_keys=True, default=str)
                elif singleton.startswith('key_'):
                    key_name = singleton[4:]
                    return json.dumps(
                        {'class': cls.__name__, 'key': kwargs.get(key_name, args[0] if args else cls.__name__)},
                        sort_keys=True, default=str)
                else:
                    return singleton
            elif isinstance(singleton, (list, tuple)):
                param_dict = {k: kwargs.get(k, '') for k in singleton}
                return json.dumps({'class': cls.__name__, 'params': param_dict}, sort_keys=True, default=str)
            return cls.__name__

        def new_ins():
            ins = type.__call__(cls, *args, **kwargs)
            for attr_name in dir(ins):
                for fd in fix.values():
                    if fd['match'](ins, attr_name):
                        fd['fix'](ins, attr_name)
            return ins

        # 检查单例
        if singleton:
            key = get_key()
            if key not in cls._instances:
                instance = new_ins()
                # 动态包装方法
                cls._instances[key] = instance
            instance = cls._instances[key]
        else:
            instance = new_ins()
        if callable(install := getattr(instance, 'install', None)):
            install()
        return instance


class Empty(metaclass=CustomMeta, singleton=True):
    def __call__(self, *args, **kwargs):
        return self

    def __bool__(self):
        return False

    def __getattr__(self, item):
        return self

    def __repr__(self):
        return "empty"

    __str__ = __repr__



# -*- coding: utf-8 -*-
"""
@File    : response.py
@Date    : 2024/6/29 下午1:57
@Author  : yintian
@Desc    :  Response Model by bricks
"""
from ytools.utils import magic

magic.require("w3lib>=2.1.2")
magic.require("lxml")

import re
import sys
from typing import Union, Callable, Any, List, Mapping

from lxml import etree
from w3lib.encoding import http_content_type_encoding, html_body_declared_encoding

from ytools.network.header import Header, Cookies
from ytools.network.request import Request

from ytools.utils import magic
from ytools.utils import extractors

_HEADER_ENCODING_RE = re.compile(r"charset=([\w-]+)", re.I)


class Response:
    def __init__(
            self,
            content: Any = None,
            status_code: int = 200,
            headers: Union[Header, dict, Mapping] = None,
            url: str = None,
            encoding: str = None,
            reason: str = 'ok',
            cookies: Cookies = None,
            history: List['Response'] = None,
            request: 'Request' = ...,
            error: Any = "",
            callback: Callable = None
    ):
        self.content: Union[str, bytes] = content
        self.status_code = status_code
        self._headers = Header(headers)
        self.url = url
        self.encoding = encoding or self.guess_encoding()
        self.reason = reason
        self.cookies = cookies
        self.history: List['Response'] = history or []
        self.request: 'Request' = request
        self.error = error
        self.callback = callback
        self._cache = {}

    headers: Header = property(
        fget=lambda self: self._headers,
        fset=lambda self, v: setattr(self, "_headers", Header(v)),
        fdel=lambda self: setattr(self, "_headers", Header({})),
        doc="请求头"
    )

    def guess_encoding(self):
        if not self.content:
            return

        # 1. 从header中获取编码
        content_type = self.headers.get("Content-Type")
        temp = http_content_type_encoding(content_type)
        if temp:
            return temp

        temp = html_body_declared_encoding(self.content)
        if temp:
            return temp

        return 'utf-8'

    @property
    def text(self):
        """
        :return: the str for response content.
        """
        if isinstance(self.content, bytes):

            if not self.content:
                return str('')

            try:
                return str(self.content, self.encoding, errors='replace')
            except (LookupError, TypeError):
                return str(self.content, errors='replace')

        else:
            return self.content

    @property
    def html(self):
        return etree.HTML(self.text)

    @property
    def length(self):
        """
        :return:the length of response content.
        """
        return len(self.content or '')

    @property
    def size(self):
        """
        :return:the size of response content in bytes.
        """
        return sys.getsizeof(self.content)

    def json(self, **kwargs):
        """
        Deserialize a JSON document to a Python object.
        """
        return magic.json_or_eval(self.text, **kwargs) if isinstance(self.text, str) else self.text

    def extract_all(
            self,
            engine: Union[str, Callable],
            rules: Union[dict, list],
    ):
        """
        根据多个规则循环匹配

        :param engine:
        :param rules:
        :return:
        """

        for rule in magic.iterable(rules):
            yield self.extract(
                engine=engine,
                rules=rule,
            )

    def extract(
            self,
            engine: Union[str, Callable],
            rules: dict,

    ):
        """
        提取引擎, 生成器模式, 支持 Rule, 批量匹配

        :param engine:
        :param rules:
        :return:
        """

        exs = {
            'JSON': extractors.JsonExtractor,
            'XPATH': extractors.XpathExtractor,
            'JASONPATH': extractors.JsonpathExtractor,
            'REGEX': extractors.RegexExtractor,
        }
        if not engine:
            return []

        try:
            if isinstance(engine, str):
                if engine.upper() in exs:
                    extractor: extractors.Extractor = exs[engine.upper()]
                    ret = extractor.match(
                        obj=self.text,
                        rules=rules
                    )
                    return ret
                else:
                    engine = magic.load_object(engine)

            if callable(engine):

                return engine({
                    'response': self,
                    'request': self.request,
                    'rules': rules,
                })

            else:
                raise ValueError(f"无法识别的引擎类型: {engine}")

        except Exception as e:
            raise RuntimeError(f"解析数据出现了意料之外的错误: {e}")

    def xpath(self, xpath, obj=None, **kwargs):
        """
        进行xpath匹配

        :param xpath: xpath规则
        :param obj: 要匹配的对象
        :param kwargs:
        :return:
        """

        if not xpath:
            return obj

        obj = obj or self.html

        return extractors.XpathExtractor.extract(
            obj=obj,
            exprs=xpath,
            **kwargs
        )

    def xpath_first(self, xpath, obj=None, default=None, **kwargs):
        """
        返回xpath的第一个对象

        :param default:
        :param xpath:
        :param obj:
        :param kwargs:
        :return:
        """

        return magic.first(self.xpath(xpath, obj, **kwargs), default=default)

    def jsonpath(self, jpath, obj=None, strict=True, **kwargs):
        """
        使用 jsonpath 进行匹配

        :param jpath:
        :param obj:
        :param strict:
        :param kwargs:
        :return:
        """
        if not jpath: return obj
        obj = obj or self.text
        return extractors.JsonpathExtractor.extract(
            obj=obj,
            exprs=jpath,
            jsonp=not strict,
            **kwargs
        )

    def jsonpath_first(self, jpath, obj=None, default=None, strict=True, **kwargs):
        """
        返回jpath的第一个对象

        :param strict:
        :param default:
        :param jpath:
        :param obj:
        :param kwargs:
        :return:
        """
        return magic.first(self.jsonpath(jpath, obj, strict, **kwargs), default=default)

    def re(self, regex, obj=None, **kwargs):
        """
        正则匹配

        :param regex: 正则表达式
        :param obj:
        :param kwargs:
        :return:
        """
        obj = obj or self.text

        if not regex:
            return obj

        return extractors.RegexExtractor.extract(
            obj=obj,
            exprs=regex,
            **kwargs
        )

    def re_first(self, regex, default=None, obj=None, **kwargs):
        """
        正则

        :param obj:
        :param regex:
        :param default:
        :param kwargs:
        :return:
        """
        return magic.first(self.re(regex, obj, **kwargs), default=default)

    def get(self, rule: str, obj=None, strict=True, **kwargs):
        """
        json匹配
        json规则示例:
        字典内直接采用 a.b.c 的方式匹配
        如果要匹配列表内的每一项, 采用$代替每一项
        如果要从父级开始匹配, 采用//代替父级节点

        :param rule: json解析规则
        :param strict: json解析规则
        :param obj: 需要匹配的dict/list对象
        :return:
        """
        obj = obj or self.text

        if not rule:
            return obj

        return extractors.JsonExtractor.extract(
            obj=obj,
            exprs=rule,
            jsonp=not strict,
            **kwargs
        )

    def get_first(self, rule: str, default=None, obj=None, strict=True, **kwargs):
        """
        返回json匹配的第一个对象

        :param rule: json解析规则
        :param strict: json解析规则
        :param default: 获取不到的时候的默认值为
        :param obj: 需要匹配的dict/list对象
        :return:

        """
        return magic.first(self.get(rule, obj, strict, **kwargs), default=default)

    @property
    def ok(self):
        """
        return true if response status_code conform to the rules

        :return:
        """
        return 200 <= self.status_code < 400

    @classmethod
    def make_response(cls, **kwargs):
        """
        make one empty response if not kwargs else normal response

        :param kwargs:
        :return:
        """
        kwargs.setdefault("status_code", 0)
        kwargs.setdefault("reason", "empty")
        kwargs.setdefault("error", "empty")
        return cls(**kwargs)

    def is_json(self, **kwargs):
        try:
            self.json(**kwargs)
            return True
        except:  # noqa
            return False

    def __str__(self):
        return f'<Response [{self.error if self.status_code == -1 else self.status_code}] {self.url}>'

    __repr__ = __str__

    def __bool__(self):
        return self.ok

    def __setattr__(self, key, value):

        if key in ["encoding", "content"]:
            # 修改这三个属性的时候, 需要把缓存清空
            object.__setattr__(self, "_cache", {})

        return object.__setattr__(self, key, value)

    def __getattribute__(self, item):

        def cache_method(func):
            def wrapper(*args, **kwargs):
                # 生成缓存的键
                cache_key = (args, tuple(kwargs.items()))
                # 检查缓存
                if cache_key not in self._cache:
                    self._cache[cache_key] = func(*args, **kwargs)
                return self._cache[cache_key]

            return wrapper

        if item in ["text", "html", "json"]:
            cache = self._cache
            if cache and item in cache:
                cached = cache[item]
                return cached
            else:
                ret = object.__getattribute__(self, item)
                self._cache[item] = cache_method(ret) if callable(ret) else ret
                return self._cache[item]

        return object.__getattribute__(self, item)


if __name__ == '__main__':
    res = Response('{"name":"kem"}')
    print(res.json())
    print(res.json())
    print(res.json())
    print(res.json())
    res.content = '{"name":"kem2"}'
    print(res.json())
    print(res.json())
    print(res.json())



# -*- coding: utf-8 -*-
"""
@File    : route.py
@Date    : 2024/8/12 上午9:17
@Author  : yintian
@Desc    : 
"""
from DrissionPage._configs.chromium_options import ChromiumOptions

from ytools.utils.magic import require

require('DrissionPage>=4.0.5.6')

import base64
import json
import random
import re
import time
from collections import defaultdict
from hashlib import md5

from DrissionPage._base.driver import Driver  # noqa
from DrissionPage._pages.chromium_page import ChromiumPage  # noqa

from ytools.utils.magic import result


class RouteRequestExtra:
    def __init__(self, **kwargs):
        self.resource_type = kwargs.get('resourceType')
        self.resource_error_reason = kwargs.get('responseErrorReason')
        self.response_status_code = kwargs.get('responseStatusCode')
        self.response_status_text = kwargs.get('responseStatusText')
        self.response_headers = kwargs.get('responseHeaders')
        # self.network_id = kwargs.get('networkId')
        self.redirect_url = kwargs.get('redirectUrl')
        self.is_download = kwargs.get('isDownload')
        self.is_navigation_request = kwargs.get('isNavigationRequest')
        self.auth_challenge = kwargs.get('authChallenge')
        self.is_download = kwargs.get('isDownload')
        self.is_download = kwargs.get('isDownload')
        request = kwargs.get('request')
        self.has_post_data = request.get('hasPostData')
        self.post_data_entries = request.get('postDataEntries')
        self.mixed_content_type = request.get('mixedContentType')
        self.initial_priority = request.get('initialPriority')
        self.referrer_policy = request.get('referrerPolicy')
        self.is_link_preload = request.get('isLinkPreload')
        self.trust_token_params = request.get('trustTokenParams')
        self.is_same_site = request.get('isSameSite')
        self.url_fragment = request.get('urlFragment')


class RouteRequest:
    response = None

    def __init__(self, **kwargs):
        self.error = None
        self.success = False
        request: dict = kwargs.get('request')
        self.interception_id = kwargs.get('interceptionId')
        self.driver: Driver = kwargs.get('driver')
        self.frame_id = kwargs.get('frameId')
        self.request_id = kwargs.get('requestId')
        self.url = request.get('url')
        self.method = request.get('method')
        self.headers = request.get('headers')
        self.post_data = request.get('postData')
        self.extra = RouteRequestExtra(**kwargs)
        self.options = kwargs.copy()
        self.options.pop('driver', None)
        self.raw_fingerprint = self.fingerprint
        self.response_headers = {}
        self.raw_data = self.continue_data.copy()

    @property
    def fingerprint(self):
        md = md5()
        md.update(json.dumps({
            **self.continue_data,
            "error": str(self.error),
            "success": self.success
        }).encode())
        return md.hexdigest()

    @property
    def continue_data(self):
        data = dict(
            interceptionId=self.interception_id,
            url=self.url,
            method=self.method,
            postData=self.post_data or '',
            headers=self.headers,
        )
        if self.response:
            body = ['HTTP/2.0 200 OK']
            for k, v in self.response_headers.items():
                body.append(f'{k}: {v}')
            body.append(f"Content-Length: {len(self.response)}")
            body.append('')
            body.append(self.response)
            response = '\r\n'.join(body)
            # body.append(base64.b64encode(self.response.encode()).decode())
            data['rawResponse'] = base64.b64encode(response.encode()).decode()
        if self.error:
            if self.error not in ['Failed', 'Aborted', 'TimedOut', 'AccessDenied', 'ConnectionClosed',
                                  'ConnectionReset', 'ConnectionRefused', 'ConnectionAborted', 'ConnectionFailed',
                                  'NameNotResolved', 'InternetDisconnected', 'AddressUnreachable',
                                  'BlockedByClient', 'BlockedByResponse']:
                self.error = 'Aborted'
            data['errorReason'] = self.error
        return data

    def fulfill_request(self, **kwargs):
        pass

    def continue_request(self, **kwargs):
        if self.raw_fingerprint == self.fingerprint:
            self.driver.run(
                'Network.continueInterceptedRequest',
                interceptionId=self.interception_id
            )
        else:
            continue_data = {
                **self.continue_data,
                **kwargs
            }
            self.driver.run(
                'Network.continueInterceptedRequest',
                **continue_data
            )
        self.success = True

    def fail_request(self, error):
        error = error or self.error
        if not error or error not in ['Failed', 'Aborted', 'TimedOut', 'AccessDenied', 'ConnectionClosed',
                                      'ConnectionReset', 'ConnectionRefused', 'ConnectionAborted', 'ConnectionFailed',
                                      'NameNotResolved', 'InternetDisconnected', 'AddressUnreachable',
                                      'BlockedByClient', 'BlockedByResponse']:
            error = 'Aborted'
        self.driver.run(
            'Fetch.failRequest',
            requestId=self.request_id,
            errorReason=error
        )
        self.error = error


class RouteResponse:
    content: bytes = b''
    encoding = 'utf-8'
    flag = False

    def __init__(self, **kwargs):
        self.success = False
        self.error = ''
        self.request: RouteRequest = kwargs.get('request')
        self.driver: Driver = kwargs.get('driver')
        self.url = self.request.url
        self.raw_fingerprint = self.fingerprint
        self.raw_data = self.continue_data.copy()
        self.headers = {}
        self.get_response_body()

    def get_response_body(self):
        if self.content:
            return self.content
        res = self.driver.run(
            'Network.getResponseBodyForInterception',
            interceptionId=self.request.interception_id
        )
        if body := res.get('body'):
            if res.get('base64Encoded'):
                self.content = base64.b64decode(body)
            else:
                if isinstance(body, str):
                    self.content = body.encode()
                elif isinstance(body, bytes):
                    self.content = body
                else:
                    raise ValueError(f'body 类型错误,应为 str, bytes, 获取到 {type(body)}')
        return self.content

    @property
    def text(self):
        return self.content.decode(encoding=self.encoding)

    @text.setter
    def text(self, value):
        if isinstance(value, str):
            self.content = value.encode(encoding=self.encoding)
        elif isinstance(value, bytes):
            self.content = value
        else:
            try:
                self.content = json.dumps(value).encode(encoding=self.encoding)
            except Exception as e:
                raise e
        self.flag = True

    def json(self):
        return json.loads(self.text)

    @property
    def continue_data(self):
        data = dict(
            interceptionId=self.request.interception_id,
            url=self.url,
            method=self.request.method,
            postData=self.request.post_data or '',
            headers=self.request.headers,
        )
        if self.flag and self.content:
            if self.headers:
                response_lines = ['HTTP/1.1 200 OK']
                # 添加默认头部
                if self.headers:
                    for key, value in self.headers.items():
                        response_lines.append(f"{key}: {value}")
                # 添加 Content-Length
                response_lines.append(f"Content-Length: {len(self.content)}")
                # 添加空行分隔头部和正文
                response_lines.append("")
                # 添加正文
                response_lines.append(self.content.decode())
                # 合并成完整的响应字符串
                full_response = "\r\n".join(response_lines)
                data['rawResponse'] = base64.b64encode(full_response.encode()).decode()
            else:
                data['rawResponse'] = base64.b64encode(self.content).decode()
            # if self.headers:
            #     data['headers'] = self.headers
        if self.error:
            if self.error not in ['Failed', 'Aborted', 'TimedOut', 'AccessDenied', 'ConnectionClosed',
                                  'ConnectionReset', 'ConnectionRefused', 'ConnectionAborted', 'ConnectionFailed',
                                  'NameNotResolved', 'InternetDisconnected', 'AddressUnreachable',
                                  'BlockedByClient', 'BlockedByResponse']:
                self.error = 'Aborted'
            data['errorReason'] = self.error
        return data

    @property
    def fingerprint(self):
        md = md5()
        md.update(json.dumps({
            **self.continue_data,
            "success": self.success
        }).encode())
        return md.hexdigest()

    def continue_request(self, **kwargs):
        if self.raw_fingerprint == self.fingerprint:
            self.driver.run(
                'Network.continueInterceptedRequest',
                interceptionId=self.request.interception_id
            )
        else:
            continue_data = {
                **self.continue_data,
                **kwargs
            }
            self.driver.run(
                'Network.continueInterceptedRequest',
                **continue_data
            )
        self.success = True


class Route:
    _patters = defaultdict(dict)
    _request = defaultdict(dict)

    @classmethod
    def start_by_driver(cls, driver: Driver):
        driver.run(
            'Network.setRequestInterception',
            patterns=[
                {
                    "urlPattern": "*",
                    "interceptionStage": "HeadersReceived"
                },
                {
                    "urlPattern": "*",
                    "interceptionStage": "Request"
                }
            ]
        )
        driver.set_callback(
            'Network.requestIntercepted',
            cls.mock_func(
                func=cls.request_pause,
                driver=driver
            )
            # cls.request_pause
        )

    @staticmethod
    def mock_func(func, *args, **kwargs):
        def inner(**k):
            _kwargs = kwargs.copy()
            _kwargs.update(k or {})
            return result(
                func=func,
                args=args,
                kwargs=_kwargs,
            )

        return inner

    @classmethod
    def set(cls, action, target, func):
        if action not in cls._patters[target]:
            cls._patters[target][action] = []
        cls._patters[target][action].append(func)

    @classmethod
    def get(cls, action, target):
        if action not in cls._patters[target]:
            cls._patters[target][action] = []
        return cls._patters[target][action]

    @classmethod
    def pop(cls, action, target, func):
        if action not in cls._patters[target]:
            cls._patters[target][action] = []
        cls._patters[target][action].remove(func)

    @classmethod
    def on(cls, action, target, func):
        cls.set(f'on_{action}', target, func)

    @classmethod
    def format_request(cls, **kwargs):
        driver: Driver = kwargs.get('driver')
        request = RouteRequest(**kwargs)
        cls._request[request.request_id]['request'] = request
        cls._request[request.request_id]['driver'] = driver
        return request

    @classmethod
    def format_response(cls, **kwargs):
        response = RouteResponse(**kwargs)
        cls._request[response.request.request_id]['response'] = response
        return response

    @classmethod
    def request_pause(cls, **kwargs):
        # driver: Driver = kwargs.get('driver')
        request = cls.format_request(**kwargs)
        if request.extra.response_status_code:
            kwargs['request'] = request
            response = cls.format_response(**kwargs)
            cls._on_response(request, response)
        else:
            cls._on_request(request)

    @classmethod
    def _on_request(cls, request: RouteRequest):
        for p, v in cls._patters.items():
            if re.search(p, request.url):
                for func in v.get('on_request', []):
                    cls.mock_func(
                        func=func,
                        request=request
                    )()
                    if request.success or request.error:
                        break
        else:
            request.continue_request()

    @classmethod
    def _on_response(cls, request: RouteRequest, response: RouteResponse):
        for p, v in cls._patters.items():
            if re.search(p, response.url):
                for func in v.get('on_response', []):
                    cls.mock_func(
                        func=func,
                        request=request,
                        response=response
                    )()
                    if request.success or request.error:
                        break
        else:
            response.continue_request()


def mock_raw(request: RouteRequest):
    request.url = 'http://106.55.2.140:5522/utils/raw?a=0&b=0'
    request.headers['t1'] = "asd"
    request.response_headers = {
        'Date': 'Tue, 11 Mar 2025 16:48:03 GMT',
        'Content-Type': 'text/plain; charset=utf-8',
        'X-Custom-Header': 'Test-Value111',
        'Access-Control-Allow-Origin': '*',  # 处理跨域
    }
    request.response = "123"
    # request.headers = {
    #     'Content-Type': 'text/plain; charset=utf-8'
    # }
    # print(kwargs)
    # request.fail_request('')


def mock_res(**kwargs):
    response: RouteResponse = kwargs['response']

    response.text = 'mock success'

    response.headers = {
        'Date': 'Tue, 11 Mar 2025 16:48:03 GMT',
        'Content-Type': 'text/plain; charset=utf-8',
        'X-Custom-Header': 'Test-Value111',
        'Access-Control-Allow-Origin': '*',  # 处理跨域
    }
    # print(kwargs)


def mock_baidu(request: RouteRequest):
    print(request.url)


def test():
    opt = ChromiumOptions()
    opt.set_argument(' --disable-web-security', "")
    opt.set_browser_path("a")
    browser = ChromiumPage(opt)
    Route.start_by_driver(driver=browser.driver)
    Route.on('request', 'utils', mock_raw)
    Route.on('response', 'utils', mock_res)
    # Route.on('response', 'baidu', mock_baidu)
    # browser.get('http://www.baidu.com')
    b = random.randint(0, 10000)
    browser.get(f'http://106.55.2.140:5522/utils/raw?a=1&b={b}')
    while True:
        time.sleep(1)


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
"""
@File    : rpa_base.py
@Date    : 2025/1/17 14:40
@Author  : yintian
@Desc    : 
"""
import re
import threading
import time
from random import uniform

from ytools import logger
from ytools.utils.magic import result


class RPAControl:
    _call_end = []
    _call_kwargs = []
    node = []

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @staticmethod
    def start_thread(target, t_args=None, t_kwargs=None, **kwargs):
        t = threading.Thread(
            target=target,
            args=t_args or (),
            kwargs=t_kwargs or {},
            **kwargs
        )
        t.start()
        return t

    def where(self) -> str:
        pass

    def to(self, tow: str, where='', default=None, force=False, call_args=None, call_kwargs=None):
        if not where:
            where = self.where()
        if not default:
            de = self.get_func(f'^to_{tow}$')
            if de:
                default = de[0]['func']
            else:
                default = lambda: logger.debug(f'当前处于 {where}, 想要去 {tow}')
        if not force and self.check_node(where, tow):
            logger.debug(f'当前为非强制前往 {tow}, 但是已经处于 {where}, 所以放弃行为')
            return
        func = default
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if callable(attr) and attr_name == f'{where}_to_{tow}':
                func = getattr(self, attr_name)
                if callable(func):
                    break
        return result(
            func=func,
            args=[
                *(call_args or [])
            ],
            kwargs={
                'where': where,
                'tow': tow,
                **(call_kwargs or {}),
                **locals()
            }
        )

    def call(self, where: str = '', default=None, suffix="", call_args=None, call_kwargs=None, try_count=3, end="", force=True):
        if end:
            if not self._call_end or (self._call_end[-1] != end):
                self._call_end.append(end)
        self._call_kwargs = {
            'where': where, 'default': default, 'suffix': suffix, 'call_args': call_args, 'call_kwargs': call_kwargs,
            'try_count': try_count, 'end': end, 'force': force,
        }
        if not default:
            default = lambda: logger.debug(
                f'当前处于 {where}' if not suffix else f'当前处于 {where}, 寻找 {suffix} 失败'
            )
        for i in range(try_count):
            if where:
                break
            where = self.where()
        if not where:
            if force:
                self.where_in_none(
                    **self._call_kwargs
                )
            else:
                logger.debug(f'当前处于深空')
            return
        if check_node := self.check_node(where):
            logger.debug(f'当前-{where} 达到或超过任务终点-{check_node}, 结束任务')
            return
        func = default
        if suffix:
            patten = f'where_in_(.*?)_for_{suffix}$'
        else:
            patten = 'where_in_(.*)$'
        for func_dict in self.get_func(patten):
            su = func_dict['su']
            su_options = su.split('_or_')
            if where in su_options:
                func = func_dict['func']
                break
        logger.debug(f'当前处于 {where}, 执行 - {func.__name__}')
        return result(
            func=func,
            args=[
                *(call_args or [])
            ],
            kwargs={
                'where': where,
                **(call_kwargs or {}),
                **locals()
            }
        )

    def check(self, *args, **kwargs):
        funcs = self.get_func('^check_for')
        log_flag = kwargs.get('log_flag', False)
        for func in funcs:
            func_name = func['func_name']
            func = func['func']
            result(
                func=func,
                args=args,
                kwargs=kwargs
            )
            log_flag and logger.debug(f'执行 {func_name}')

    def clear_call(self):
        self._call_end = []
        self._call_kwargs = {}

    def confused(self, where='', confused_pre='', force=False, *args, **kwargs):
        confused_type = '_'.join(filter(bool, [confused_pre, 'confused'], ))
        self.call(where=where, suffix=confused_type, call_args=args, call_kwargs=kwargs, force=force)

    def get_func(self, name_patten):
        res = []
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if not callable(attr):
                continue
            su = re.findall(name_patten, attr_name)
            if not su:
                continue
            res.append({"func_name": attr_name, "func": attr, "su": su[0]})
        return res

    def check_node(self, now_node, _check_node=None):
        if now_node not in self.node:
            return False
        if not _check_node and not self._call_end:
            return False
        check_node = _check_node or self._call_end[-1]
        for node in self.node:
            if check_node == node:
                not _check_node and self._call_end.pop(-1)
                return check_node
            if now_node == node:
                break
        return False

    def run(self):
        self.call()

    @staticmethod
    def deal_error(error, error_cls=Exception, error_type='ignore'):
        if error_type == 'ignore':
            logger.error(error)
        else:
            if isinstance(error, Exception):
                raise error
            raise error_cls(error)

    @staticmethod
    def sleep(t1=1.0, t2=1.0):
        time.sleep(t1 + uniform(0, t2))

    def where_in_none(self, *args, **kwargs):
        logger.debug(f'当前位置无法识别')
        time.sleep(30)
        kwargs['end'] = ''
        self.call(*args, **kwargs)


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : track.py
@Date    : 2025/1/17 15:10
@Author  : yintian
@Desc    : 
"""

# -*- coding: utf-8 -*-

# @File    : track.py
# @Date    : 2023-03-09 14:32
# @Author  : yintian
# @Desc    :

import json
import re
import subprocess
import threading
import time
from queue import Queue
from random import choice, randint

from loguru import logger

lock = threading.Lock()

track_queue = Queue()

GLOBAL_INFO = {}


def put2queue(track_list):
    track_queue.put(track_list)


def iter_queue():
    while True:
        if not track_queue.empty():
            if item := track_queue.get():
                yield [item]


def listener(
        filename=None,
        save=False,
        plat='pc',
        min_length=100,
        interval=0.001,
        **kwargs
):
    """
    录制总接口
    :param filename: 存储的文件名
    :param save: 是否存储到本地文件
    :param plat: 录制平台
    :param min_length: 最小录制长度
    :param interval: 录制间隔
    :param kwargs:
    :return:
    """
    assert bool(save) != bool(filename), 'save 和 filename 只需要一个且不能同时存在'

    def save():
        for data in iter_queue():
            save_data(
                data=data,
                filename=filename,
                min_length=min_length,
                **kwargs
            )

    threading.Thread(target=save, daemon=True).start()

    if plat == 'pc':
        listen_pc(interval)
    elif plat == 'android':
        listen_android(**kwargs)
    else:
        raise ValueError(f'Platform {plat} not supported')


def listen_android(
        device='/dev/input/event2',
        window_size=None,
        max_xy=None
):
    """
     监听 android 的轨迹
    :param device: 通过 adb shell getevent -lp 找到和屏幕相关的设备, 可以输出后复制给 GPT 帮忙找
    :param window_size: 可不传
    :param max_xy: 可不传
    :return:
    """
    xy_pattern = re.compile(r'(\d+\.\d+).*ABS_MT_POSITION_([XY])\s+(\w+)')
    flag_pattern = re.compile(r'(\d+\.\d+).*BTN_TOUCH\s+(\w+)')

    def get_screen_resolution():
        """获取手机屏幕分辨率"""
        res = subprocess.run(['adb', 'shell', 'wm', 'size'], capture_output=True, text=True)
        resolution = re.search(r'(\d+)x(\d+)', res.stdout)
        if resolution:
            width, height = map(int, resolution.groups())
            return width, height
        else:
            raise Exception("无法获取屏幕分辨率")

    def get_max_xy():
        """获取手机屏幕分辨率"""
        res = subprocess.run('adb shell dumpsys input'.split(' '), capture_output=True, text=True)
        resolution = re.findall(r'([XY]): min=0, max=(\d+)', res.stdout)
        max_x, max_y = 0, 0
        for r in resolution:
            if r[0] == 'X':
                max_x = int(r[1])
            elif r[0] == 'Y':
                max_y = int(r[1])
            if max_x and max_y:
                return max_x, max_y
        raise Exception("无法获取最大像素值")

    def get_radio():
        sx, sy = window_size or get_screen_resolution()
        max_x, max_y = max_xy or get_max_xy()
        return sx / max_x, sy / max_y

    radio_x, radio_y = get_radio()
    process = subprocess.Popen(
        ['adb', 'shell', 'getevent', '-lt', device],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True
    )
    flag = False
    x, y, t = 0, 0, 0
    x0, y0, t0 = 0, 0, 0
    track_list = []
    try:
        for line in process.stdout:
            if flag_match := re.findall(flag_pattern, line):
                ts, action = flag_match[0]
                if action == 'DOWN':
                    flag = True
                elif action == 'UP':
                    logger.debug(track_list)
                    put2queue(track_list.copy())
                    track_list.clear()
                    x0, y0, t0 = 0, 0, 0
                    x, y, t = 0, 0, 0
                    flag = False
            if not flag:
                continue
            xy_match = re.findall(xy_pattern, line)
            if not xy_match:
                continue
            ts, xy, value_hex = xy_match[0]
            value = int(value_hex, 16)
            if xy == 'X':
                x = value
                t = int(float(ts) * 1000)
            elif xy == 'Y':
                if t != int(float(ts) * 1000):
                    x = 0
                    continue
                y = value
            if all([x, y, t]):
                if not any([x0, y0, t0]):
                    x0, y0, t0 = x, y, t
                track_list.append(((x - x0) * radio_x, (y - y0) * radio_y, t - t0))
                x, y, t = 0, 0, 0
    except KeyboardInterrupt:
        pass
    finally:
        # 确保进程结束
        process.terminate()


def listen_pc(
        interval=0.001
):
    try:
        from pynput.mouse import Listener
    except ImportError:
        logger.error(
            '请安装 pynput 库 -- 参考指令: pip install -i https://pypi.tuna.tsinghua.edu.cn/simple pynput==1.7.6')
        return

    # 可变化的原点
    x0, y0, t0 = 0, 0, 0
    # 录制标志
    flag = False
    # 临时队列
    res = []

    def on_move(x, y):
        nonlocal t0
        now = int(time.time() * 1000)
        if t0 == 0:
            t0 = now
        res.append((int(x - x0), int(y - y0), int(time.time() * 1000) - t0))
        time.sleep(interval)

    def on_click(x, y, button, pressed):  # noqa
        nonlocal flag, res, x0, y0, t0
        flag = not flag
        logger.debug('开始录制' if flag else '结束录制')
        if flag:
            x0, y0, t0 = x, y, 0
            res.clear()
        else:
            now = res.copy()
            logger.info(now)
            put2queue(now)

    with Listener(on_move=on_move, on_click=on_click) as listen:
        listen.join()


def save_data(data, filename=None, min_length=100, save_mode='override'):
    data = [_ for _ in data if len(_) > min_length]
    if not data:
        logger.debug(f'未有符合条件的轨迹,不保存')
        return
    if filename:
        logger.debug(f'保存历史轨迹至文件 {filename}: {json.dumps(data)}')
        if save_mode == 'override':
            with open(filename, 'w', encoding='utf-8') as f:
                json.dump(data, f)  # noqa
        else:
            try:
                with open(filename, 'r', encoding='utf-8') as f:
                    data = json.load(f)
            except Exception:  # noqa
                data = []
            data.extend(data)
            with open(filename, 'w', encoding='utf-8') as f:
                json.dump(data, f)  # noqa


def get_track(
        distance: int,
        track_list: list = None,
        max_dis=1000,
        file_name="",
        min_node=100,
        tag=None
):
    """
    获取随机轨迹
    :param distance: 需生成的轨迹长度, 需要小于 max_dis 以及 track_list中的最小长度的三分之一, 用以保证轨迹的随机性
    :param track_list: 自带轨迹, 在该轨迹中随机选取一段轨迹
    :param max_dis: 限制最大传入长度
    :param file_name: 需要读取的文件名
    :param min_node: 最少的节点数, 用以保证轨迹的随机性
    :param tag: 
    :return:
    """

    def get_track_list() -> list:
        """
        获取 track_list 信息

        :return: track_list 信息
        """
        try:
            with open(file_name, 'r', encoding='utf-8') as f:
                _track_list = json.load(f)
            return _track_list
        except Exception as e:
            raise RuntimeError(f'[请求失败] 失败原因: 获取 track_list 信息 失败，失败信息：{str(e)}')

    def binary_search(lis: list, x, left=None, right=None, ):
        """
        二分查找
        :param lis: 被查找的队列
        :param x: 目标值
        :param left: 起始点
        :param right: 结束点
        :return:
        """
        left = left or 0
        right = right or len(lis)
        while left <= right:
            mid = left + (right - left) // 2
            if lis[mid] == x:
                return mid
            elif lis[mid] > x:
                right = mid - 1
            else:
                left = mid + 1
        return right

    def get_one_track():
        """
        获取一条轨迹
        :return:
        """
        one_track_xy = choice(track_list)
        if len(one_track_xy) < min_node:
            logger.debug(f'轨迹节点数小于 {min_node}, 重新获取轨迹')
            return get_one_track()
        one_track = [_[0] for _ in one_track_xy]
        start_index = randint(0, len(one_track) - 2)
        last = one_track[-1]
        start = one_track[start_index]
        return last, start, one_track, one_track_xy, start_index

    def get_target_track():
        """
        获取目标轨迹
        :return:
        """
        last, start, one_track, one_track_xy, start_index = get_one_track()
        while last - start < distance:
            last, start, one_track, one_track_xy, start_index = get_one_track()
        target = start + distance
        target_index = binary_search(one_track, target)
        _res = one_track_xy[start_index: target_index + 2]
        start_row = _res[0]
        _result = []
        for _ in _res:
            _result.append([_[i] - start_row[i] for i in range(len(_))])
        _distance = target - start_row[0]
        while _result[-2][0] >= _distance:
            _result.pop(-1)
        _result[-1][0] = target - start_row[0]
        return _result

    while not (track_list or GLOBAL_INFO[f'$track_list_{tag}']):
        # 只有拿到锁之后才会去获取 track_list 指纹
        if lock.acquire():
            try:
                GLOBAL_INFO[f'$track_list_{tag}'] = get_track_list()
            finally:
                lock.release()
        else:
            logger.debug('等待 track_list 获取')
            time.sleep(5)

    track_list = track_list or [json.loads(_['track']) for _ in GLOBAL_INFO[f'$track_list_{tag}']]

    assert track_list != [], "未获取到符合条件的 track_list !"

    if distance > max_dis:
        raise ValueError('Distance 太大了')

    track = get_target_track()
    return track


if __name__ == '__main__':
    # print(get_track(130, tag='default'))
    listener(filename='a.json', tag='low1', plat='pc')

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : rpa.py
@Date    : 2025/1/17 15:02
@Author  : yintian
@Desc    : 
"""
import shutil
import time
import typing
from random import randint

from DrissionPage._configs.chromium_options import ChromiumOptions  # noqa
from DrissionPage._elements.chromium_element import ChromiumElement  # noqa
from DrissionPage._pages.chromium_page import ChromiumPage  # noqa
from DrissionPage._pages.chromium_tab import ChromiumTab  # noqa

from ytools import logger
from ytools.auto_driver.dp.route import Route
from ytools.auto_driver.rpa_base import RPAControl
from ytools.auto_driver.track import get_track
from ytools.utils.magic import single, iterable


class DpRpaBase(RPAControl):
    _page: ChromiumPage = None
    _tab: ChromiumTab = None
    options: ChromiumOptions = None
    route = Route
    _local_path: str = None
    listen_at = None

    def __init__(self, **kwargs):
        self.set_options(kwargs.get('options_kwargs', {}), **kwargs.get('options_others', {}))
        super().__init__(**kwargs)

    def goto(self, url, page=None, **kwargs):
        kwargs.setdefault('show_errmsg', False)
        kwargs.setdefault('retry', 0)
        kwargs.setdefault('interval', None)
        kwargs.setdefault('timeout', 30)

        page = page or self.page
        page.get(url, **kwargs)

    def set_options(self, options: dict = None, **kwargs):
        options = options or {}
        op = ChromiumOptions()
        for k, v in options.items():
            op.set_argument(k, v if v else None)
        if server := kwargs.get('server'):
            if isinstance(server, int):
                op.set_local_port(server)
            elif isinstance(server, str):
                if server.isdigit():
                    op.set_local_port(server)
                else:
                    op.set_address(server)
            else:
                raise ValueError
        if local_path := kwargs.get('user_data_path', self._local_path):
            op.set_user_data_path(local_path)
            self._local_path = local_path
        self.options = op
        if self._page:
            del self._page

    @property
    def page(self):
        if not self._page and not self._tab:
            self._page = ChromiumPage(self.options)
        if self.listen_at and not self._page.listen.listening:
            self._page.listen.start(self.listen_at)
        return self._tab or self._page

    def set_page(self, page: typing.Optional[ChromiumPage]):
        if not page:
            self.set_tab(None)
        self._page = page

    def set_tab(self, tab: typing.Optional[ChromiumTab]):
        self._tab = tab

    def re_tab(self):
        try:
            logger.debug(f'关闭当前标签, 启动新标签')
            self.close_other_tab()
            self.goto('about:blank')
        except Exception as e:
            logger.error(f're_tab-{e}')

    def close_other_tab(self):
        self.page.close_tabs([_ for _ in self.page.tab_ids if _ != self.page.tab_id])

    def close(self, rm=True):
        try:
            if self._tab:
                self._tab.close()
            if self._page:
                self._page.browser.quit(force=True)
            self.set_page(None)
            time.sleep(5)
        except Exception as e:
            logger.error(f'关闭浏览器失败, {e}', )
        try:
            rm and shutil.rmtree(self._local_path)
        except Exception as e:
            logger.error(f'删除浏览器本期文件失败, {e}')

    def find(self, eles: typing.Union[typing.List[typing.Union[str, dict]], str, dict], index=0, page: ChromiumPage = None, one=True, error_type='ignore', try_count=3) \
            -> typing.Union[typing.List[ChromiumElement], ChromiumElement]:
        def find_one(_ele, _index, parent=None):
            try:
                parent = parent or page
                r = parent.eles(_ele, timeout=1)
                if len(r) < _index:
                    self.deal_error(f'匹配到对应元素: {_ele},但 index-{index} 超标, 匹配长度为 {len(r)}',
                                    error_type=error_type)
                    return None
                r = r[_index] if one and r and _index >= 0 else r
                return r if not one else single(r)
            except Exception as e:
                self.deal_error(e, error_type=error_type)
            return None

        def find_dict(_ele: dict):
            try:
                if shadow_root := _ele.get('shadow_root'):
                    parent = page.ele(shadow_root).shadow_root
                else:
                    parent = page
                r = find_one(_ele.get('ele'), _ele.get('index', 0), parent=parent)
                r = r[index] if r and index > 0 else r
                return one and r
            except Exception as e:
                self.deal_error(e, error_type=error_type)
            return None

        res = []
        page = page or self.page
        for ele in iterable(eles):
            for _ in range(try_count):
                if isinstance(ele, str):
                    _r = find_one(ele, index)
                elif isinstance(ele, list) and len(ele) > 1:
                    _r = find_one(ele[0], ele[1])
                elif isinstance(ele, dict):
                    _r = find_dict(ele)
                else:
                    _r = ele
                if _r and one:
                    return single(_r)
                elif _r:
                    res.extend(list(iterable(_r)))
                    break
        return res  # noqa

    def run_slider(self, distance: int, slider_ele='#sliderContainer'):
        slider = single(self.find(slider_ele))
        track = self.get_track_list(
            distance=distance
        )
        self.page.actions.hold(slider)
        now = slider.rect.midpoint
        for t in track:
            self.page.actions.move_to(now, t[0], t[1], duration=0.006)
        self.sleep(0.1, 0.2)
        self.page.actions.release()

    @staticmethod
    def get_track_list(distance):
        def get_one_track(dis, file_name=""):
            while True:
                track = get_track(distance=dis, file_name=file_name)
                if len(track) > 2:
                    break
            return track

        while True:
            fast_dis = randint(distance // 2, distance * 2 // 3)
            low_dis = randint(12, 20)
            middle_dis = distance - fast_dis - low_dis
            if middle_dis > 0:
                break
        fast_track = get_one_track(fast_dis, 'fast.json')
        middle_track = get_one_track(middle_dis, 'middle.json')
        low_track = get_one_track(low_dis, 'low.json')
        _res = fast_track
        last_node = _res[-1]
        for _ in middle_track[1:]:
            _res.append([_[0] + last_node[0], _[1] + last_node[1]], )
        last_node = _res[-1]
        for _ in low_track[1:]:
            _res.append([_[0] + last_node[0], _[1] + last_node[1]])
        _res.append([distance, _res[-1][1]])
        return _res

    def click(self, eles: typing.Union[typing.List[typing.Union[str, dict]], str, dict], index=0, page: ChromiumPage = None, one=True, error_type='ignore', try_count=3, sleep=3):
        def click_one(_ele):
            try:
                if r := self.find(_ele, index=index, page=page, one=True, error_type=error_type, try_count=1):
                    return any([button.click() for button in iterable(r)])
                self.sleep()
            except Exception as e:
                self.deal_error(e, error_type=error_type)
            finally:
                self.sleep(sleep)

        page = page or self.page
        for ele in iterable(eles):
            flag = False
            for _ in range(try_count):
                flag = click_one(ele)
                if flag:
                    break
            if flag and one:
                break

    def input(self, eles: typing.Union[typing.List[typing.Union[str, dict]], str, dict], text: str, index=0, page: ChromiumPage = None, one=True, error_type='ignore', try_count=3):

        def input_one(_ele, _text, _index=0):
            try:
                if r := self.find(_ele, index=index, page=page, one=True, error_type=error_type, try_count=1):
                    if isinstance(_ele, list) and len(_ele) > 2:
                        clear = bool(_ele[2])
                    elif isinstance(_ele, dict):
                        clear = bool(_ele.get('clear'))
                    else:
                        clear = True
                    if isinstance(_ele, list) and len(_ele) > 3:
                        by_js = bool(_ele[3])
                    elif isinstance(_ele, dict):
                        by_js = bool(_ele.get('by_js'))
                    else:
                        by_js = False
                    [button.input(_text, clear=clear, by_js=by_js) for button in iterable(r)]
                    return True
                self.sleep()
            except Exception as e:
                self.deal_error(e, error_type=error_type)

        page = page or self.page
        for ele in iterable(eles):
            flag = False
            for _ in range(try_count):
                flag = input_one(ele, text, index)
                if flag:
                    break
            if flag and one:
                break

    def quit(self):
        if self._page:
            self._page.close()
        self._page = None  # noqa
        self._tab = None  # noqa

    def where_in_none(self, *args, **kwargs):
        logger.debug(f'当前位置无法识别')
        logger.debug(f'当前 url: {self.page.url}')
        time.sleep(30)
        kwargs['end'] = ''
        self.call(*args, **kwargs)

    def __del__(self):
        self.close()


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : header.py
@Date    : 2024/6/29 下午1:57
@Author  : yintian
@Desc    :  Header mode by bricks
"""
import typing
from collections import UserDict
from http.cookiejar import CookieJar
from http.cookies import SimpleCookie
from typing import Mapping

CookieTypes = typing.Union[
    "Cookies", CookieJar, typing.Dict[str, str], typing.List[typing.Tuple[str, str]]
]


# Code stolen from curl_cffi.requests.cookies.Cookies
class Cookies(typing.MutableMapping[str, str]):
    """
    HTTP Cookies, as a mutable mapping.
    """

    def __init__(self, cookies: typing.Optional[CookieTypes] = None) -> None:
        if cookies is None or isinstance(cookies, dict):
            self.jar = CookieJar()
            if isinstance(cookies, dict):
                for key, value in cookies.items():
                    self.set(key, value)
        elif isinstance(cookies, list):
            self.jar = CookieJar()
            for key, value in cookies:
                self.set(key, value)
        elif isinstance(cookies, Cookies):
            self.jar = CookieJar()
            for cookie in cookies.jar:
                self.jar.set_cookie(cookie)
        else:
            self.jar = cookies

    def set(self, name: str, value: str, domain: str = "", path: str = "/") -> None:
        """
        Set a cookie value by name. May optionally include domain and path.
        """
        kwargs = {
            "version": 0,
            "name": name,
            "value": value,
            "port": None,
            "port_specified": False,
            "domain": domain,
            "domain_specified": bool(domain),
            "domain_initial_dot": domain.startswith("."),
            "path": path,
            "path_specified": bool(path),
            "secure": False,
            "expires": None,
            "discard": True,
            "comment": None,
            "comment_url": None,
            "rest": {"HttpOnly": None},
            "rfc2109": False,
        }
        cookie = Cookie(**kwargs)  # type: ignore
        self.jar.set_cookie(cookie)

    def get(  # type: ignore
            self,
            name: str,
            default: typing.Optional[str] = None,
            domain: typing.Optional[str] = None,
            path: typing.Optional[str] = None,
    ) -> typing.Optional[str]:
        """
        Get a cookie by name. May optionally include domain and path
        in order to specify exactly which cookie to retrieve.
        """
        value = None
        matched_domain = ""
        for cookie in self.jar:
            if cookie.name == name:
                if domain is None or cookie.domain == domain:
                    if path is None or cookie.path == path:
                        # if cookies on two different domains do not share a same value
                        if (
                                value is not None
                                and not matched_domain.endswith(cookie.domain)
                                and not str(cookie.domain).endswith(matched_domain)
                                and value != cookie.value
                        ):
                            message = (
                                f"Multiple cookies exist with name={name} on "
                                f"{matched_domain} and {cookie.domain}"
                            )
                            raise ValueError(message)
                        value = cookie.value
                        matched_domain = cookie.domain or ""

        if value is None:
            return default
        return value

    def delete(
            self,
            name: str,
            domain: typing.Optional[str] = None,
            path: typing.Optional[str] = None,
    ) -> None:
        """
        Delete a cookie by name. May optionally include domain and path
        in order to specify exactly which cookie to delete.
        """
        if domain is not None and path is not None:
            return self.jar.clear(domain, path, name)

        remove = [
            cookie
            for cookie in self.jar
            if (
                    cookie.name == name
                    and (domain is None or cookie.domain == domain)
                    and (path is None or cookie.path == path)
            )
        ]

        for cookie in remove:
            self.jar.clear(cookie.domain, cookie.path, cookie.name)

    def clear(
            self, domain: typing.Optional[str] = None, path: typing.Optional[str] = None
    ) -> None:
        """
        Delete all cookies. Optionally include a domain and path in
        order to only delete a subset of all the cookies.
        """
        args = []
        if domain is not None:
            args.append(domain)
        if path is not None:
            assert domain is not None
            args.append(path)
        self.jar.clear(*args)

    def update(self, cookies: typing.Optional[CookieTypes] = None) -> None:  # noqa
        cookies = Cookies(cookies)
        for cookie in cookies.jar:
            self.jar.set_cookie(cookie)

    def __setitem__(self, name: str, value: str) -> None:
        return self.set(name, value)

    def __getitem__(self, name: str) -> str:
        value = self.get(name)
        if value is None:
            raise KeyError(name)
        return value

    def __delitem__(self, name: str) -> None:
        return self.delete(name)

    def __len__(self) -> int:
        return len(self.jar)

    def __iter__(self) -> typing.Iterator[str]:
        return (cookie.name for cookie in self.jar)

    def __bool__(self) -> bool:
        for _ in self.jar:
            return True
        return False

    def __repr__(self) -> str:
        cookies_repr = ", ".join(
            [
                f"<Cookie {cookie.name}={cookie.value} for {cookie.domain} />"
                for cookie in self.jar
            ]
        )

        return f"<Cookies[{cookies_repr}]>"

    @classmethod
    def by_jar(cls, jar):
        cookies = cls()
        for cookie in jar:
            if isinstance(cookie, dict):
                cookies.set(name=cookie["name"], value=cookie["value"], domain=cookie.get('domain'),
                            path=cookie.get('path'))
            else:
                cookies.set(name=cookie.name, value=cookie.value, domain=cookie.domain, path=cookie.path)
        return cookies

    def load(self, set_cookie_string: str):
        simple_cookie = SimpleCookie()
        simple_cookie.load(set_cookie_string)
        # 遍历解析后的 SimpleCookie 来创建 Cookie 对象并添加到 CookieJar
        for key, morsel in simple_cookie.items():
            self.set(name=key, value=morsel.value, domain=morsel["domain"], path=morsel["path"])


class Header(UserDict):

    def __init__(self, data=None, **kwargs):
        self.caches = {}
        if data is None:
            data = {}
        super().__init__({**data, **kwargs})

    def __setitem__(self, key, value):
        if key.lower() in self.caches:
            return super().__setitem__(self.caches[key.lower()], value)
        else:
            self.caches[key.lower()] = key
            return super().__setitem__(key, value)

    def __getitem__(self, key):

        if key.lower() in self.caches:
            return super().__getitem__(self.caches[key.lower()])
        else:
            return super().__getitem__(key)

    def __delitem__(self, key):
        if key.lower() in self.caches:
            return super().__delitem__(self.caches[key.lower()])
        else:
            return super().__delitem__(key)

    def __eq__(self, other):
        if isinstance(other, (Mapping, dict)):
            other = self.__class__(other)
        else:
            return NotImplemented
        # Compare insensitively
        return self.items() == other.items()

    def __contains__(self, __o: object) -> bool:
        if isinstance(__o, str):
            key = __o.lower()
        else:
            key = __o

        return key in self.caches


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : request.py
@Date    : 2024/6/29 下午1:57
@Author  : yintian
@Desc    :  Request mode by bricks
"""
import argparse
import re
import shlex
import urllib.parse
from collections import OrderedDict
from http.cookies import SimpleCookie
from typing import Union, Optional, Dict, List, Callable

from ytools.network.header import Header


class Request:

    def __init__(
            self,
            url: str,
            params: Optional[dict] = None,
            method: str = 'GET',
            data: Union[str, dict] = None,
            json_data: Union[dict, list] = None,
            headers: Union[Header, dict] = None,
            cookies: Dict[str, str] = None,
            options: dict = None,
            timeout: int = ...,
            allow_redirects: bool = True,
            proxies: Optional[str] = None,
            proxy: Optional[Union[dict, str, List[Union[dict, str]]]] = None,
            ok: Optional[Union[str, Dict[str, Union[Callable]]]] = ...,
            retry: int = 0,
            max_retry: Union[int, float] = 5,
            use_session: bool = False,
    ) -> None:
        """

        :param url: 请求 URL
        :param params: 请求 URL 参数
        :param method: 请求方法, 默认 GET
        :param headers: 请求头
        :param cookies: 请求 cookies
        :param options: 请求其他额外选项, 可以用于配合框架 / 下载器
        :param timeout: 请求超时时间, 填入 ... 为默认
        :param allow_redirects: 是否允许重定向
        :param proxies: 请求代理 Optional[str], 如 http://127.0.0.1:7890, 理解为 current proxy
        :param proxy: 代理 Key, 理解为 proxy from
        :param ok: 判断成功动态脚本, 字符串形式, 如通过 403 状态码可以写为: 200 <= response.status_code < 400 or response.status_code == 403
        :param retry: 当前重试次数
        :param max_retry: 最大重试次数
        :param use_session: 是否使用下载器的 session 模式
        """
        self.url = url
        self.use_session = use_session
        self.params = params
        self.cookies = cookies
        self.method = method.upper()
        self.data = data
        self.json = json_data
        self._headers = Header(headers)
        self.options = options or {}
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.proxies = proxies
        self.proxy: Optional[Union[dict, str, List[Union[dict, str]]]] = proxy
        self.ok = ok
        self.retry = retry
        self.max_retry = max_retry

    headers: Header = property(
        fget=lambda self: self._headers,
        fset=lambda self, v: setattr(self, "_headers", Header(v)),
        fdel=lambda self: setattr(self, "_headers", Header({})),
        doc="请求头"
    )

    @property
    def real_url(self):
        # 解析原始URL
        parsed_url = urllib.parse.urlparse(self.url)
        # 提取查询字符串并解析为字典
        original_params = dict(urllib.parse.parse_qsl(parsed_url.query))

        # 更新原始参数字典
        original_params.update(self.params or {})

        # 重新构建查询字符串
        query_string = urllib.parse.urlencode(original_params).replace('+', '%20')

        # 构建新的URL
        new_url = urllib.parse.urlunparse((
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            parsed_url.params,
            query_string,
            parsed_url.fragment
        ))

        return new_url

    @property
    def curl(self):
        # 开始构建curl命令
        parts = ["curl", f"-X {self.method.upper()}"]

        # 添加请求头

        if self.cookies:
            cookie_str = "; ".join([f'{k}={v}' for k, v in self.cookies.items()])
        else:
            cookie_str = ""

        content_type = ''
        if self.headers:
            for header, value in self.headers.items():
                if header.lower() == 'cookie':
                    cookie_str = "; ".join(list(filter(None, [value, cookie_str])))
                else:
                    if header.lower() == 'content-type':
                        content_type = value.lower()
                    parts.append(f"-H {shlex.quote(f'{header}: {value}')}")

        cookie_str and parts.append(f"-H {shlex.quote(f'Cookie: {cookie_str}')}")

        if self.json:
            body = self.json
        elif self.data:
            body = self.data.decode() if isinstance(self.data, bytes) else self.data
        else:
            body = None
        # 添加请求体
        if body:
            if content_type.__contains__('application/json'):
                # JSON格式的请求体
                parts.append(f"-d {shlex.quote(body)}")
            elif content_type.__contains__('application/x-www-form-urlencoded'):
                # URL编码格式的请求体
                form_data = urllib.parse.urlencode(body)
                parts.append(f"--data {shlex.quote(form_data)}")
            else:
                if isinstance(body, dict):
                    body_data = urllib.parse.urlencode(body)
                else:
                    body_data = body
                # 其他或未知类型，假定为字符串
                parts.append(f"--data {shlex.quote(body_data)}")

        # 添加URL（包括查询参数）
        parts.append(shlex.quote(self.real_url))

        # 转换为字符串格式的命令
        return " ".join(parts)

    @classmethod
    def from_curl(cls, curl_command):
        """
        从 curl 中导入

        :param curl_command:
        :return:
        """

        _parser = argparse.ArgumentParser()
        _parser.add_argument('command')
        _parser.add_argument('url')
        _parser.add_argument('-d', '--data')
        _parser.add_argument('-c', '--cookie', default=None)
        _parser.add_argument('-r', '--request', default=None)
        _parser.add_argument('-p', '--proxy', default=None)
        _parser.add_argument('-b', '--data-binary', '--data-raw', default=None)
        _parser.add_argument('-X', default='')
        _parser.add_argument('-H', '--header', action='append', default=[])
        _parser.add_argument('-du', '--data-urlencode', action='append', default=[], type=lambda x: x.split("="))
        _parser.add_argument('--compressed', action='store_true')
        _parser.add_argument('--location', action='store_true')
        _parser.add_argument('-k', '--insecure', action='store_true')
        _parser.add_argument('--user', '-u', default=())
        _parser.add_argument('-i', '--include', action='store_true')
        _parser.add_argument('-s', '--silent', action='store_true')

        if isinstance(curl_command, str):
            curl_command = curl_command.replace('curl --location', 'curl')
            tokens = shlex.split(curl_command.replace(" \\\n", " "))
        else:
            tokens = curl_command

        parsed_args = _parser.parse_args(tokens)

        if parsed_args.data_urlencode:
            post_data = {i[0]: i[1] for i in parsed_args.data_urlencode}
        else:
            post_data = parsed_args.data or parsed_args.data_binary

        method = parsed_args.request.lower() if parsed_args.request else 'post' if post_data else "get"

        if parsed_args.X:
            method = parsed_args.X.lower()

        cookie_dict = OrderedDict()
        if parsed_args.cookie:
            cookie = SimpleCookie(bytes(parsed_args.cookie, "ascii").decode("unicode-escape"))
            for key in cookie:
                cookie_dict[key] = cookie[key].value

        quoted_headers = OrderedDict()

        for curl_header in parsed_args.header:
            if curl_header.startswith(':'):
                occurrence = [m.start() for m in re.finditer(':', curl_header)]
                header_key, header_value = curl_header[:occurrence[1]], curl_header[occurrence[1] + 1:]
            else:
                header_key, header_value = curl_header.split(":", 1) if ':' in curl_header else (curl_header, "")

            if header_key.lower().strip("$") == 'cookie':
                cookie = SimpleCookie(bytes(header_value, "ascii").decode("unicode-escape"))
                for key in cookie:
                    cookie_dict[key] = cookie[key].value
            else:
                quoted_headers[header_key] = str(header_value).strip()

        # 解析原始URL
        parsed_url = urllib.parse.urlparse(parsed_args.url)
        # 提取查询字符串并解析为字典
        params = dict(urllib.parse.parse_qsl(parsed_url.query))
        if (quoted_headers.get('Content-Type') or quoted_headers.get('content-type')).__contains__("json"):
            data = {"json": post_data}
        else:
            data = {"data": post_data}

        return cls(
            url=parsed_url.scheme + "://" + parsed_url.netloc + parsed_url.path,
            params=params,
            method=method,
            headers=quoted_headers,
            cookies=cookie_dict,
            **data
        )

    def put_options(self, key: str, value, action="update"):
        func = getattr(self.options, action)
        return func(**{key: value})

    def get_options(self, key: str, default=None, action="get"):
        func = getattr(self.options, action)
        return func(key, default)

    def __str__(self):
        return f'<Request [{self.real_url}]>'

    __repr__ = __str__


if __name__ == '__main__':
    # req = Request(
    #     url="https://example.com/path",
    #     params={"query": "value"},
    #     method="POST",
    #     body="key=value",
    #     # headers={"Content-Type": "application/json"}
    # )
    # print(req.curl)
    req = Request.from_curl(
        """
       curl --location 'https://api.m.jd.com/api' \
--header 'Host: api.m.jd.com' \
--header 'Connection: keep-alive' \
--header 'X-Referer-Page: /pages/product/product' \
--header 'xweb_xhr: 1' \
--header 'X-Rp-Client: mini_2.0.0' \
--header 'User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36 MicroMessenger/6.8.0(0x16080000) NetType/WIFI MiniProgramEnv/Mac MacWechat/WMPF MacWechat/3.8.5(0x1308050b)XWEB/31023' \
--header 'X-Referer-Package: wx862d8e26109609cb' \
--header 'Accept: */*' \
--header 'Sec-Fetch-Site: cross-site' \
--header 'Sec-Fetch-Mode: cors' \
--header 'Sec-Fetch-Dest: empty' \
--header 'Referer: https://servicewechat.com/wx862d8e26109609cb/134/page-frame.html' \
--header 'Accept-Language: zh-CN,zh;q=0.9' \
--header 'Content-Type: application/x-www-form-urlencoded' \
--data-urlencode 'appid=jd_healthy' \
--data-urlencode 'functionId=jdh_mini_viewHealthy' \
--data-urlencode 'body={"t":"1699945509078","fromType":"wxapp","apolloId":"0bffab44e51a45dd9abba25bb1a837d4","apolloSecret":"63776250363d4f58be246942e1d7dffc","clientVersion":"6.0.0","moudleId":"product","source":7,"wareId":"10035789319416","pageParams":{"wareId":"10032049399319"}}' \
--data-urlencode 'client=jdh_android' \
--data-urlencode 'clientVersion=191' \
--data-urlencode 'uuid=c916c23b-3f47-44df-9676-991a4572d9b7' \
--data-urlencode 'h5st=20231114150509103;39nitmmzz55zz5t9;3c4d4;tk03a89ae1bce18pMngxKzF4M3gx-8IEi7HovD4yyS7t9CS44Pf87FKEDnwhVM_flAfOnMwpsj0nCM527xQUv_NvUQIs;7169fa348bca9b8d1a2f69abcbe9a69c;4.1;1699945509103;275420328beca586eb4d58ea10bc056f1633a9c9f91541c365a5c67afa45389b1b104ca161b54546027decda1169ec06e8acaf7230ba14f851bd4921d74e865e63ccb34bc0b1c584632a2e22048ed0158b238cf99ee17e8a7bde6196c7a66692' \
--data-urlencode 'x-api-eid-token=jdd01w49LHCK5MMEEHEICAQSQJG7A5GGN6I3JZ3QOV5JYACPJPDQ5VWLWCTG3LJZW2LZOE336QNRLTNC6FOMNDCWIEUDXRD6VN2H2BPKLEACLMCUVOPELA5XV2DMG3GWB6MPTYIT'
        """
    )

    print(req.curl)



# -*- coding: utf-8 -*-
"""
@File    : date.py
@Date    : 2024/4/16 下午8:19
@Author  : yintian
@Desc    : 
"""
import time
from collections.abc import Callable
from typing import Literal

from ytools.utils.magic import require

require("arrow==1.3.0", action="fix")

from arrow.constants import DEFAULT_LOCALE  # noqa
from arrow import Arrow as _Arrow, get as arrow_get  # noqa

T_FRAMES = Literal[
    "year",
    "years",
    "month",
    "months",
    "day",
    "days",
    "hour",
    "hours",
    "minute",
    "minutes",
    "second",
    "seconds",
    "microsecond",
    "microseconds",
    "week",
    "weeks",
    "quarter",
    "quarters",
]


class Arrow(_Arrow):

    @classmethod
    def get(
            cls,
            *args,
            **kwargs
    ) -> 'Arrow':
        """
        获取 Arrow 对象
        :param args:
        :param kwargs:
        :return:
        """
        return cls.fromdatetime(arrow_get(*args, **kwargs).datetime)

    def ts(self, length=10, fmt=int):
        """
        获取时间戳

        :param length: 长度为几位
        :param fmt: 返回什么类型
        :return:
        """
        return fmt(self.timestamp() * 10 ** (length - 10))

    def start(
            self,
            frame: T_FRAMES
    ) -> 'Arrow':
        """
        获取当天开始时间
        :return:
        """
        return self.get(self.floor(frame))

    def end(
            self,
            frame: T_FRAMES
    ) -> 'Arrow':
        """
        获取结束的时间
        :return:
        """
        return self.get(self.ceil(frame))

    def spend(self, fmt: Callable = None):
        t = time.time() - self.ts(10)
        if fmt:
            t = fmt(t)
        return t

    def format(
            self, fmt: str = "YYYY-MM-DD HH:mm:ss", locale: str = DEFAULT_LOCALE
    ) -> str:
        return super().format(fmt, locale)


if __name__ == '__main__':
    a = Arrow.now()
    print(a.format())



# -*- coding: utf-8 -*-
"""
@File    : error.py
@Date    : 2024-06-03 15:35
@Author  : yintian
@Desc    : 
"""


class ResultError(Exception):

    def __init__(self, error, stack, f, a=None, k=None):
        super().__init__()
        self.error = error
        self.stack = stack
        self.func = f
        self.args = a or tuple()
        self.kwargs = k or dict()

    def __str__(self):
        return f'<{self.error.__class__.__name__}: func:{self.func}, args:{self.args}, kwargs:{self.kwargs}>'

    def __bool__(self):
        return False


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : version.py
@Author  : yintian
@Date    : 2025/6/23 14:02
@Software: PyCharm
@Desc    : 
"""
import os

from ytools.utils.magic import require

require("packaging")

from packaging.version import Version


def get_version(package='ytools', show=False, path=__file__):
    this_dir = os.path.dirname(os.path.dirname(path))
    version_path = os.path.join(this_dir, package, 'VERSION')
    if not os.path.exists(version_path):
        print(f'未找到 VERSION 文件')
        return
    with open(version_path, 'r') as f:
        version = f.read()
    show and print(version)
    return version


def update_version(version=None, update_type='auto', package='ytools', save=True, path=__file__):
    # 读取当前版本
    version = version or get_version(package=package, path=path)
    try:
        v = Version(version)
    except:  # noqa
        raise ValueError(f"无效的版本号格式: {version}")

    # 提取版本组件
    major = v.major
    minor = v.minor
    micro = v.micro
    pre = v.pre  # 预发布版本 (例如: ('b', 1))
    dev = v.dev  # 开发版本
    post = v.post  # 后发布版本
    epoch = v.epoch  # 纪元版本

    if update_type == 'auto' and pre:
        update_type = 'pre'

    # 根据更新类型处理版本
    if update_type in ('micro', 'auto'):
        micro += 1
        pre = post = dev = None
    elif update_type == 'minor':
        minor += 1
        micro = 0
        pre = post = dev = None
    elif update_type == 'major':
        major += 1
        minor = micro = 0
        pre = post = dev = None
    elif update_type.startswith('pre'):
        if pre:
            pre = (pre[0], pre[1] + 1)
        else:
            pre = (update_type[4:] if len(update_type) > 4 else 'a', 1)  # 默认创建alpha版本
            micro += 1
    else:
        raise ValueError(f"无效的更新类型: {update_type}")

    # 构建新版本号
    new_version = Version(f"{epoch}!{major}.{minor}.{micro}" +
                          (f"{pre[0]}{pre[1]}" if pre else "") +
                          (f".post{post}" if post else "") +
                          (f".dev{dev}" if dev else "")).public
    save and save_version(new_version, package, path=path)
    return new_version


def save_version(version, package='ytools', path=__file__):
    this_dir = os.path.dirname(os.path.dirname(path))
    version_path = os.path.join(this_dir, package, 'VERSION')
    # 写回文件
    with open(version_path, 'w') as f:
        f.write(str(version) + '\n')


def is_main_version(version):
    try:
        v = Version(version)
    except:  # noqa
        raise ValueError(f"无效的版本号格式: {version}")
    return not any([v.post, v.dev, v.pre, v.epoch])


if __name__ == '__main__':
    update_version(save=False)



# -*- coding: utf-8 -*-
"""
@File    : magic.py
@Date    : 2024/4/16 下午8:17
@Author  : yintian
@Desc    : 
"""
import ast
import importlib
import inspect
import json
import os
import re
import subprocess
import sys
from collections import namedtuple
from typing import Literal, Any, List, Union, Generator, Dict, Callable
from urllib import parse

import better_exceptions
import importlib_metadata
from loguru import logger

from ytools.error import ResultError
from ytools.utils.magic_for_class import Empty

better_exceptions.MAX_LENGTH = None
exec_formatter = better_exceptions.ExceptionFormatter(
    colored=False,
    theme=better_exceptions.THEME,
    max_length=better_exceptions.MAX_LENGTH,
    pipe_char=better_exceptions.PIPE_CHAR,
    cap_char=better_exceptions.CAP_CHAR
)

PACKAGE_REGEX = re.compile(r"([a-zA-Z0-9_\-]+)([<>=]*)(.*)")

# 一般而言, 是一个以 simple 结尾的 url
PIPY_REGEX = re.compile(r"https?://.*/simple/?$")
PYPI_MIRROR = {
    "TUNA": "https://pypi.tuna.tsinghua.edu.cn/simple",
    "USTC": "http://pypi.mirrors.ustc.edu.cn/simple/",
    "Aliyun": "http://mirrors.aliyun.com/pypi/simple/",
    "Tencent": "https://mirrors.cloud.tencent.com/pypi/simple/",
    "Huawei": "https://repo.huaweicloud.com/repository/pypi/simple/",
    "pypi": "https://pypi.org/simple/"
}


def require(
        package_spec: str,
        action: Literal["raise", "fix"] = "fix",
        mirror_sources: str = "TUNA",
        kwargs: Dict[str, str] = None,
        mode=None
):
    kwargs = kwargs or {}
    mode = mode or os.environ.get('$require_mode') or 'pip'
    match mode:
        case "pip":
            require_pip(package_spec, action, mirror_sources, kwargs)
        case "uv":
            require_uv(package_spec, action, mirror_sources, kwargs)
        case _:
            if callable(func := locals().get('require_uv')):
                func(package_spec, action, mirror_sources, kwargs)
            else:
                raise ValueError(f"can not find require_{mode}")


def require_uv(
        package_spec: str,
        action: Literal["raise", "fix"] = "fix",
        mirror_sources: str = "TUNA",
        uv_kwargs: Dict[str, str] = None
):
    package_spec = re.sub(r"\s+", "", package_spec)
    match = PACKAGE_REGEX.match(package_spec)
    mirror_sources = PYPI_MIRROR.get(mirror_sources, mirror_sources)

    assert match, ValueError(f"无效的包规范: {package_spec}")
    assert not mirror_sources or PIPY_REGEX.match(mirror_sources), ValueError(f"无效的镜像源: {mirror_sources}")
    package, operator, required_version = match.groups()
    mirror_sources and uv_kwargs.setdefault('--index', mirror_sources)

    try:
        import uv  # noqa
    except ImportError:
        require("uv", mode="pip", action="fix")
        import uv  # noqa

    try:
        # 获取已安装的包版本
        installed_version = importlib_metadata.version(package)
        # 检查是否需要安装或更新
        from ytools.utils.package import parse as version_parse
        if required_version and not eval(
                f'version_parse({installed_version!r}) {operator} version_parse({required_version!r})'):
            raise importlib_metadata.PackageNotFoundError
        else:
            return installed_version

    except importlib_metadata.PackageNotFoundError:

        # 包没有安装或版本不符合要求
        install_command = package_spec if required_version else package
        cmd = [uv.find_uv_bin(), "add"]
        for k, v in uv_kwargs.items():
            k and cmd.append(k)
            v and cmd.append(v)
        cmd.append(install_command)
        if action == "raise":
            raise importlib_metadata.PackageNotFoundError(f"依赖包不符合要求, 请使用以下命令安装: {' '.join(cmd)}")
        else:
            logger.debug(f"依赖包不符合要求, 自动修正, 命令: {' '.join(cmd)}")
            subprocess.check_call(cmd)
            return importlib_metadata.version(package)


def require_pip(
        package_spec: str,
        action: Literal["raise", "fix"] = "fix",
        mirror_sources: str = "TUNA",
        pip_kwargs: Dict[str, str] = None
) -> str:
    """
    依赖 python 包

    :param action: 依赖操作, 若为 fix 则自动下载对应包,
    :param package_spec: pymongo==4.6.0 / pymongo
    :param mirror_sources : pip 源, 内置有 "TUNA", "USTC", "Aliyun", "Tencent", "Huawei", "pypi", 可以传自己的源进来
    :param pip_kwargs: pip 的一些参数, 如 --trusted-host, 单独的参数可以使用 {"--upgrade": ""} 这种方式
    :return: 安装的包版本号
    """
    # 分离包名和版本规范
    package_spec = re.sub(r"\s+", "", package_spec)
    match = PACKAGE_REGEX.match(package_spec)
    mirror_sources = PYPI_MIRROR.get(mirror_sources, mirror_sources)
    pip_kwargs = pip_kwargs or {}

    assert match, ValueError(f"无效的包规范: {package_spec}")
    assert not mirror_sources or PIPY_REGEX.match(mirror_sources), ValueError(f"无效的镜像源: {mirror_sources}")
    mirror_sources and pip_kwargs.setdefault('-i', mirror_sources)
    package, operator, required_version = match.groups()

    try:
        # 获取已安装的包版本
        installed_version = importlib_metadata.version(package)
        # 检查是否需要安装或更新
        from ytools.utils.package import parse as version_parse  # noqa
        if required_version and not eval(
                f'version_parse({installed_version!r}) {operator} version_parse({required_version!r})'):
            raise importlib_metadata.PackageNotFoundError
        else:
            return installed_version

    except importlib_metadata.PackageNotFoundError:

        # 包没有安装或版本不符合要求
        install_command = package_spec if required_version else package
        cmd = [sys.executable, "-m", "pip", "install"]
        for k, v in pip_kwargs.items():
            k and cmd.append(k)
            v and cmd.append(v)
        cmd.append(install_command)
        if action == "raise":
            raise importlib_metadata.PackageNotFoundError(f"依赖包不符合要求, 请使用以下命令安装: {' '.join(cmd)}")
        else:
            logger.debug(f"依赖包不符合要求, 自动修正, 命令: {' '.join(cmd)}")
            subprocess.check_call(cmd)
            return importlib_metadata.version(package)


def load_object(path, reload=False, from_path=False, strict=True, __env__=None):
    """
    加载给定绝对对象路径的对象，并返回它。
    对象可以是类、函数、变量或实例的导入路径，
    如果 ``path`` 不是字符串，而是可调用对象，
    例如类或函数，则按原样返回。

    :param __env__:
    :param strict:
    :param from_path: 从绝对路径中加载
    :param reload: 是否重载模块
    :param path: 模块的路径
    :return:

    """

    __env__ = __env__ or {**globals(), **locals()}

    def _load(o):

        try:
            return eval(o, __env__)
        except:  # noqa
            try:
                _ret = importlib.import_module(o)
                reload and inspect.ismodule(_ret) and importlib.reload(_ret)
                return _ret

            except Exception as e:
                if strict:
                    raise ValueError(f"加载对象失败: {raw}")
                else:
                    return e

    if not isinstance(path, str):
        return path

    if from_path:
        base_dir = os.getcwd()
        if path.__contains__(base_dir):
            sons = path.replace(base_dir, "").split(".")[0].replace("/", ".")
        else:
            i = 0
            for i, s in enumerate(path):
                try:
                    assert base_dir[i] == s
                except AssertionError:
                    sons = path[i:].split(".")[0].replace("/", ".")
                    break
            else:
                sons = os.path.basename(path.split(".")[0])
        spec2 = importlib.util.spec_from_file_location(sons, path)  # noqa
        odm = importlib.util.module_from_spec(spec2)  # noqa
        spec2.loader.exec_module(odm)
        return odm
    else:
        if not re.search(r'[a-zA-Z.]+', path):
            return path

        raw = path
        # 获取调用参数
        if path.__contains__('?'):
            path, para = path.split('?', 1)
            args, kwargs = [], string2argv(para)
        else:
            args, kwargs = [], {}
        if re.match(r'lambda .*?', path):
            parent, sons, attrs = path, "", ""
        else:
            # 拆分模块与方法
            if path.__contains__(":"):
                path, attrs = path.split(":")
            else:
                path, attrs = path, ""

            parent, sons = path.rsplit(".", 1) if path.__contains__(".") else (path, "")
        # 导入父级模块
        ret = _load(parent)
        if isinstance(ret, Exception):
            return raw
        # 导入子集模块
        for attr in filter(bool, [*sons.split("."), *attrs.split(".")]):
            try:
                ret = getattr(ret, attr)
            except:  # noqa
                if strict:
                    raise ValueError(f"加载对象失败: {raw}")
                else:
                    return raw
        else:
            if raw.__contains__('?') and callable(ret):
                return result(
                    func=ret,
                    args=args,
                    kwargs=kwargs,
                    strict=True
                )
            else:
                return ret


class Prepare:

    def __init__(self, func: Union[str, Callable], args=None, kwargs=None, annotations=None, namespace=None):
        if isinstance(func, str):
            func = load_object(func, strict=True)
        self.func = func
        self.args: List = list(args) if args else []
        self.kwargs: dict = kwargs or {}
        self.parameters = inspect.signature(self.func).parameters
        self.annotations = annotations or {}
        self.namespace = namespace or {}

    def __getattr__(self, name):
        return getattr(self.func, name)

    def set_kwargs(self, key, value, force=False):
        if not self.parameters.get(key):
            return
        if force:
            self.kwargs[key] = value
        else:
            self.namespace[key] = value

    def re_build(self, *args, **kwargs):
        kwargs['namespace'] = {**self.namespace, **kwargs.get('namespace', {})}
        kwargs['annotations'] = {**self.annotations, **kwargs.get('annotations', {})}
        pre = prepare(self.func, *self.args, *args, **kwargs)
        self.__dict__ = pre.__dict__

    def __call__(self, *args, **kwargs):
        (args or kwargs) and self.re_build(*args, **kwargs)

        def run():
            def run_async():
                import asyncio

                async def _run():
                    return await self.func(*self.args, **self.kwargs)

                loop = asyncio.get_event_loop()
                fu = asyncio.run_coroutine_threadsafe(_run(), loop)
                return fu.result()

            def run_sync():
                return self.func(*self.args, **self.kwargs)

            if inspect.iscoroutinefunction(self.func) or inspect.isawaitable(self.func):
                return run_async()
            else:
                return run_sync()

        return run()

    def __repr__(self):
        return f"<{self.__class__.__name__} func: {self.func}>"

    def __eq__(self, other):
        return self.func == other


def prepare(
        func: Callable,
        *args,
        annotations: dict = None,
        namespace: dict = None,
        **kwargs,
) -> Prepare:
    annotations = annotations or {}
    namespace = namespace or {}
    args = list(args or [])
    kwargs = kwargs or {}
    positional_only = []
    positional_or_keyword = []
    var_positional = []
    keyword_only = {}
    var_keyword = {}
    signature = inspect.signature(func)
    sig_param = signature.parameters

    def get_positional(k, s):
        if args:
            v = args.pop(0)
        elif k in kwargs:
            v = kwargs.pop(k)
        else:
            v = namespace.get(k, annotations.get(k, s.default))
        return v

    def get_keyword(k, s):
        v = kwargs.pop(k, namespace.get(k, annotations.get(k, s.default)))
        return v

    for key, sig in sig_param.items():
        value = sig.empty
        match sig.kind.name:
            case 'POSITIONAL_ONLY':
                value = get_positional(key, sig)
                positional_only.append(value)
            case 'POSITIONAL_OR_KEYWORD':
                value = get_positional(key, sig)
                positional_or_keyword.append(value)
            case 'VAR_POSITIONAL':
                value = args
                var_positional = value
            case 'KEYWORD_ONLY':
                value = get_keyword(key, sig)
                keyword_only[key] = value
            case 'VAR_KEYWORD':
                value = kwargs
                var_keyword = value
        if value == sig.empty:
            raise TypeError(f"缺少参数: {key}, 参数列表: {dict(sig_param)}")
    return Prepare(
        func=func,
        args=(*positional_only, *positional_or_keyword, *var_positional),
        kwargs={**keyword_only, **var_keyword}
    )


def result(
        func,
        args: Union[list, tuple] = None,
        kwargs: dict = None,
        strict: bool = True,
        debug: bool = True,
        annotations: dict = None,
        namespace: dict = None,
        to_generator: bool = False,
        to_gen_kwargs: dict = None
):
    """
    运行一个函数并获取结果, 可以自动修复参数错误, 移除无效参数; 补充缺失参数

    :param annotations:
    :param namespace:
    :param to_gen_kwargs:
    :param to_generator:
    :param debug:
    :param strict:
    :param func: 需要调用的函数引用或相对位置
    :param args: 函数的位置参数
    :param kwargs: 函数的关键字参数
                   - _fix_return: 是否自动修正返回值
                   - _strict: 是否抛出异常
    :return:
    """
    func = load_object(func)
    args = args or tuple()
    kwargs = kwargs or {}
    assert callable(func), ValueError(f'func 必须是可执行的方法, func: {func}')
    try:
        pre = prepare(func, *args, **kwargs, namespace=namespace, annotations=annotations)
        _result = pre()
    except Exception as e:
        if not strict:
            debug and logger.exception(e)
            _result = ResultError(
                error=e,
                stack=fmt_stack(e),
                f=func,
                a=args,
                k=kwargs
            )
        else:
            raise
    return _result if to_generator is False else generator(_result, **(to_gen_kwargs or {}))


def iterable(_object: Any, enforce=(dict, str, bytes), exclude=(), convert_null=True) -> List[Any]:
    """
    用列表将 `_exclude` 类型中的其他类型包装起来

    :rtype: object
    :param convert_null: 单个 None 的时候是否给出空列表
    :param exclude: 属于此类型,便不转换
    :param enforce: 属于这里的类型, 便强制转换, 不检查 iter, 优先级第一
    :param _object:
    :return:
    """
    if _object is None and convert_null:
        return []
    if isinstance(_object, enforce) and not isinstance(_object, exclude):
        return [_object]
    elif isinstance(_object, exclude) or hasattr(_object, "__iter__"):
        return _object
    else:
        return [_object]


def generator(_object: Any, **kwargs) -> Generator:
    if inspect.isgenerator(_object):
        return _object
    elif inspect.isasyncgen(_object):
        return _object  # noqa
    else:
        kwargs.setdefault("enforce", object)
        return (i for i in iterable(_object, **kwargs))


def to_list(_object, **kwargs) -> List:
    return [*iterable(_object, **kwargs)]


def single(_object, default=None):
    """
    将元素变为可迭代对象后, 获取其第一个元素

    :param _object:
    :param default:
    :return:
    """
    return next((i for i in iterable(_object)), default)


def first(_object, default=None):
    """
    将元素变为可迭代对象后, 获取其第一个元素

    :param _object:
    :param default:
    :return:
    """
    return next((i for i in iterable(_object)), default)


def guess(_object: Any) -> Any:
    """
    智能的转换类型

    :param _object:
    :return:
    """
    if isinstance(_object, dict):
        return {k: guess(v) for k, v in _object.items()}
    elif isinstance(_object, (list, tuple, set)):
        return _object.__class__([guess(i) for i in _object])
    elif isinstance(_object, str):
        try:
            return ast.literal_eval(_object)
        except:  # noqa
            return _object
    else:
        return _object


def json_or_eval(text, jsonp=False, errors="strict", _step=0, **kwargs) -> Union[dict, list, str]:
    """
    通过字符串获取 python 对象，支持json类字符串和jsonp字符串

    :param _step:
    :param jsonp: 是否为 jsonp
    :param errors: 错误处理方法
    :param text: 需要解序列化的文本
    :return:

    """

    def literal_eval():
        return ast.literal_eval(text)

    def json_decode():
        return json.loads(text)

    def use_jsonp():
        real_text = re.search('\S+?\((?P<obj>[\s\S]*)\)', text).group('obj')
        return json_or_eval(real_text, jsonp=True, _step=_step + 1, **kwargs)

    if not isinstance(text, str):
        return text

    funcs = [json_decode, literal_eval]
    jsonp and _step == 0 and funcs.append(use_jsonp)
    for func in funcs:
        try:
            return func()
        except:  # noqa
            pass
    else:
        if errors != 'ignore':
            raise ValueError(f'illegal json string: `{text}`')
        else:
            return text


def string2argv(_string: str):
    if not isinstance(_string, str):
        return _string

    options = {}
    if not _string:
        return options

    try:
        options = json_or_eval(_string)
        assert isinstance(options, dict)
        return options
    except:  # noqa
        _string = parse.unquote(_string)
        return {
            j[0].strip(): guess(j[1].strip()) for i in _string.strip().split("&") if i for j in [i.split("=", 1)] if
            "=" in i
        }


def fmt_stack(e: Exception):
    """
    获取 stack 信息

    """
    return "".join(list(exec_formatter.format_exception(e.__class__, e, sys.exc_info()[2])))


def make_origin(*obj: Any):
    from time import time
    from random import random
    from hashlib import md5
    id_str = '-'.join([str(id(_)) for _ in obj])
    return md5(f'{id_str}-{time()}-{random()}'.encode()).hexdigest()


def check_func_or_method(obj: object):
    """

    :param obj:
    :return:
    """
    res = namedtuple('result', ["type", "cls"])

    if inspect.ismethod(obj):
        return res("method", obj.__self__, )
    else:
        if '.' in (qn := obj.__qualname__):
            obj_repr: str = qn.rsplit('.', 1)[0]
            if not obj_repr.endswith('<locals>'):
                return res("cls_method", f"{obj.__module__}.{obj_repr}")
        return res("function", None)


empty = Empty()
if __name__ == '__main__':
    require('curl-cffi>=0.5.10', action='raise')



# -*- coding: utf-8 -*-
"""
@File    : nodes.py
@Author  : yintian
@Date    : 2025/6/24 11:38
@Software: PyCharm
@Desc    : 
"""
import dataclasses
from collections.abc import Callable
from itertools import count
from types import NoneType
from typing import Self, List


@dataclasses.dataclass
class LinkNode:
    root: Callable = None
    parent: Self = None
    next: Self = None
    callback: Callable = None

    def __setattr__(self, key, value):
        if key == 'next' and type(value) not in (LinkNode, NoneType):
            value = LinkNode(value, self)
        elif key == 'next' and isinstance(value, LinkNode):
            value.parent = self

        return super().__setattr__(key, value)

    def __str__(self):
        return f"<LinkNode({self.root})>"

    __repr__ = __str__


tour_id = count()


@dataclasses.dataclass
class TourTree:
    root: Callable = None
    parent: Self = None
    child: List[Self] = dataclasses.field(default_factory=list)
    is_done: bool = False
    callback: Callable = None
    id: int = dataclasses.field(default_factory=lambda: next(tour_id))
    __used: count = dataclasses.field(default_factory=lambda: count(1))
    _used = 0

    def __setattr__(self, key, value):
        if key == 'child' and value is None:
            self.child.clear()
            return
        return super().__setattr__(key, value)

    def add(self, child):
        if child is None:
            return
        if isinstance(child, TourTree):
            child.parent = self
        else:
            child = TourTree(child, self)
        self.child.append(child)

    def done(self):
        self.is_done = True

    def tour_one(self):
        for one in self.child:
            if not one.is_done:
                return one
        if not self.parent and not self.is_done:
            return self
        if self.parent and (one := self.parent.tour_one()) and not one.is_done:
            return one

    def tour(self):
        one = self
        while one:
            yield one
            one.done()
            one = one.tour_one()

    def use(self):
        self._used = next(self.__used)

    @property
    def used(self):
        return self._used

    def print_tree(self, level=0, is_last=False, prefix='', node_info_func=None):
        """可视化打印当前节点及其子树结构"""
        # 当前节点信息
        info = f"Node {self.id} (used={self.used}, done={self.is_done})" if not node_info_func else node_info_func(self)

        # 根节点特殊处理
        if level == 0:
            print(info)
        else:
            # 根据是否最后一个子节点选择连接符
            connector = '└── ' if is_last else '├── '
            print(f"{prefix}{connector}{info}")

        # 递归打印子节点
        for i, child in enumerate(self.child):
            is_last_child = i == len(self.child) - 1
            new_prefix = prefix + ('    ' if is_last else '│   ')
            child.print_tree(level + 1, is_last_child, new_prefix, node_info_func=node_info_func)

    def __len__(self):
        return len(self.child)

    def __bool__(self):
        return bool(self.root or self.child)

    def __call__(self, *args, **kwargs):
        self.use()
        return self.root(*args, **kwargs)

    def __repr__(self):
        return f"<{self.__class__.__name__} id:{self.id} used:{self.used} parent:{self.parent.id if self.parent else None} root:{self.root}>"


if __name__ == '__main__':
    # 根节点（id=0）
    rt = lambda x: print(x)
    root = TourTree(root=rt)

    # 第一层子节点（直接挂在根节点下）
    root.add(TourTree(root=rt))  # id=1
    root.add(TourTree(root=rt))  # id=2
    root.add(TourTree(root=rt))  # id=5

    # 第二层子节点（挂在 id=1 的节点下）
    root.child[0].add(TourTree(root=rt))  # id=3
    root.child[0].add(TourTree(root=rt))  # id=4

    # 第二层子节点（挂在 id=5 的节点下）
    root.child[2].add(TourTree(root=rt))  # id=6
    root.child[2].add(TourTree(root=rt))  # id=7
    # 第三层子节点（挂在 id=7 的节点下）
    root.child[2].child[1].add(TourTree(root=rt))  # id=8
    r9 = TourTree(root=rt)
    root.child[2].child[1].add(r9)  # id=9


    def print_tree(node: TourTree, level=0):
        print(f"{'  ' * level}Node {node.id}-{node.used}-{node.is_done}")
        for child in node.child:
            print_tree(child, level + 1)

    for i in r9.tour():
        i(i)
        if i.id == 5:
            for _ in range(3):
                i.add(TourTree(root=rt))
    root.print_tree()



# -*- coding: utf-8 -*-
"""
@File    : quiter.py
@Date    : 2024/5/23 上午9:47
@Author  : yintian
@Desc    : 
"""
import atexit
import os
import signal
from typing import Callable

from ytools.log import logger
from ytools.utils.magic import make_origin, result


class Quiter:
    events = []

    @classmethod
    def run(cls, a, b):  # NOQA
        try:
            cls.run_ev()
        finally:
            logger.debug('收到退出信号, 准备退出程序')
            os._exit(0)  # noqa

    @classmethod
    def run_ev(cls):
        while cls.events:
            ev = cls.events.pop()
            try:
                result(
                    ev['func'],
                    args=ev['args'],
                    kwargs=ev['kwargs']
                )
            except:  # noqa
                pass


try:
    signal.signal(signalnum=signal.SIGTERM, handler=Quiter.run)
    signal.signal(signalnum=signal.SIGINT, handler=Quiter.run)
    signal.signal(signalnum=signal.SIGKILL, handler=Quiter.run)
except:  # noqa
    pass


def at_exit(func: Callable, *args, **kwargs):
    if kwargs.pop('__always', None):
        atexit.register(func, *args, **kwargs)
    origin = make_origin(func)
    func_dict = {
        'func': func,
        'args': args,
        'kwargs': kwargs,
        'key': origin
    }
    Quiter.events.append(func_dict)
    return origin


def un_exit(key):
    events = Quiter.events.copy()
    for ev in events:
        if ev['key'] == key:
            func = ev['func']
            Quiter.events.remove(ev)
            atexit.unregister(func)
            break



# -*- coding: utf-8 -*-
"""
@File    : __init__.py.py
@Date    : 2024/4/15 下午7:28
@Author  : yintian
@Desc    : 
"""

from ytools.log import logger
from ytools.version import get_version
from ytools.utils.magic import empty

__version__ = get_version(path=__file__)

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : encrypt.py
@Author  : yintian
@Date    : 2025/7/23 16:39
@Software: PyCharm
@Desc    : 
"""
import base64
import json
import sys
from hashlib import md5 as raw_md5

from ytools.utils.magic import json_or_eval

default_encoding = sys.getdefaultencoding()


def md5(obj, encoding=default_encoding):
    if not isinstance(obj, (str, bytes)):
        obj = json.dumps(obj, default=str, indent=0).encode(encoding=encoding)
    if isinstance(obj, str):
        obj = obj.encode(encoding=encoding)

    return raw_md5(obj).hexdigest()


def to_b64(obj, encoding=default_encoding):
    if not isinstance(obj, (str, bytes)):
        obj = json.dumps(obj, default=str, indent=0).encode(encoding=encoding)
    if isinstance(obj, str):
        obj = obj.encode(encoding=encoding)
    return base64.b64encode(obj).decode(encoding=encoding)


def from_b64(obj, fmt=True, encoding=default_encoding):
    obj = base64.b64decode(obj)
    if callable(fmt):
        return callable(obj)
    elif fmt:
        if encoding:
            obj = obj.decode(encoding=encoding)
        return json_or_eval(obj)
    return obj


if __name__ == '__main__':
    print(md5('123'))



# -*- coding: utf-8 -*-
"""
@File    : __init__.py.py
@Date    : 2024-04-17 13:14
@Author  : yintian
@Desc    : 
"""

from .response import Response
from .header import Header, Cookies
from .request import Request

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
@File    : file.py
@Date    : 2024/4/15 下午7:29
@Author  : yintian
@Desc    : 
"""

from os import path


def get_file_read(filename, base_dir='', encoding='utf-8', mode='r', safe=False):
    file_path = path.join(base_dir, filename)
    if not path.exists(file_path):
        if safe:
            return ''
        raise FileNotFoundError(f'File not found: {file_path}')
    with open(file_path, encoding=encoding, mode=mode) as f:
        long_description = f.read()
    return long_description


if __name__ == '__main__':
    pass



# -*- coding: gbk -*-
"""
@File    : setup.py
@Date    : 2024/4/15 狟敁7:23
@Author  : yintian
@Desc    : 
"""
# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

from ytools.utils.file import get_file_read
from ytools.log import logger
from ytools.version import update_version, save_version

# Package meta-data.
NAME = 'why-tools'
DESCRIPTION = '峈妦繫遜猁妏蚚梗腔馱撿儸ˋ'
URL = 'https://github.com/yintian710/whytools'
EMAIL = 'yintian710@gmail.com'
AUTHOR = 'yintian'
REQUIRES_PYTHON = '>=3.8.0'

VERSION = update_version(save=False)

# What packages are required for this module to be executed?
REQUIRED = [
    line.strip() for line in get_file_read('requirements.txt', safe=True).splitlines()
    if not line.startswith('#')
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {
    '__version__': VERSION
}


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def log(s):
        logger.info(s)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        if len(sys.argv) > 1:
            return
        try:
            self.log('Removing previous builds＃')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.log('Building Source and Wheel (universal) distribution＃')
        os.system(f'{sys.executable} setup.py sdist bdist_wheel --universal')
        self.log('Building Source and Wheel (universal) distribution＃')
        os.system(f'{sys.executable} setup.py sdist build')

        self.log('Uploading the package to PyPI via Twine＃')
        os.system('twine upload dist/*')
        save_version(version=about['__version__'])

        self.log('Pushing git tags＃')
        os.system('git commit -a -m v{0}'.format(about['__version__']))
        os.system('git push')
        # os.system()
        if 'b' not in about['__version__']:
            os.system('git tag v{0}'.format(about['__version__']))
            os.system('git push --tags')
        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=[*find_packages(exclude=["test", "demo", "old"])],
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    package_data={
        'ytools': ["VERSION"]
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
    script_name="setup.py",
    script_args=["upload"] if len(sys.argv) == 1 else sys.argv[1:],
    keywords=['ytools']
)

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

# @File    : track.py
# @Date    : 2023-03-09 14:32
# @Author  : yintian
# @Desc    :

import json
import os
import re
import subprocess
import sys
import threading
import time
from collections import defaultdict
from queue import Queue
from random import choice, randint

from loguru import logger
from pynput.keyboard import Key

from ytools.utils import magic
from ytools.utils.magic import require

lock = threading.Lock()

track_queue = Queue()
local_track = defaultdict(list)


def put2queue(track_list):
    track_queue.put(track_list)


def iter_queue():
    while True:
        if not track_queue.empty():
            if item := track_queue.get():
                yield [item]


def listener(
        filename=None,
        save=False,
        tag='default',
        plat='pc',
        min_length=100,
        interval=0.001,
        **kwargs
):
    """
    录制总接口
    :param filename: 存储的文件名
    :param save: 存储函数
    :param tag: 录制标签
    :param plat: 录制平台
    :param min_length: 最小录制长度
    :param interval: 录制间隔
    :param kwargs:
    :return:
    """
    assert bool(save) != bool(filename), 'save 和 filename 只需要一个且不能同时存在'

    def save():
        for data in iter_queue():
            save_data(
                data=data,
                filename=filename,
                save=save,
                tag=tag,
                min_length=min_length,
                interval=interval,
                **kwargs
            )

    threading.Thread(target=save, daemon=True).start()

    if plat == 'pc':
        listen_pc(interval)
    elif plat == 'android':
        listen_android(**kwargs)
    else:
        raise ValueError(f'Platform {plat} not supported')


def listen_android(
        device='/dev/input/event2',
        window_size=None,
        max_xy=None
):
    """
     监听 android 的轨迹
    :param device: 通过 adb shell getevent -lp 找到和屏幕相关的设备, 可以输出后复制给 GPT 帮忙找
    :param window_size: 可不传
    :param max_xy: 可不传
    :return:
    """
    xy_pattern = re.compile(r'(\d+\.\d+).*ABS_MT_POSITION_([XY])\s+(\w+)')
    flag_pattern = re.compile(r'(\d+\.\d+).*BTN_TOUCH\s+(\w+)')

    def get_screen_resolution():
        """获取手机屏幕分辨率"""
        res = subprocess.run(['adb', 'shell', 'wm', 'size'], capture_output=True, text=True)
        resolution = re.search(r'(\d+)x(\d+)', res.stdout)
        if resolution:
            width, height = map(int, resolution.groups())
            return width, height
        else:
            raise Exception("无法获取屏幕分辨率")

    def get_max_xy():
        """获取手机屏幕分辨率"""
        res = subprocess.run('adb shell dumpsys input'.split(' '), capture_output=True, text=True)
        resolution = re.findall(r'([XY]): min=0, max=(\d+)', res.stdout)
        max_x, max_y = 0, 0
        for r in resolution:
            if r[0] == 'X':
                max_x = int(r[1])
            elif r[0] == 'Y':
                max_y = int(r[1])
            if max_x and max_y:
                return max_x, max_y
        raise Exception("无法获取最大像素值")

    def get_radio():
        sx, sy = window_size or get_screen_resolution()
        max_x, max_y = max_xy or get_max_xy()
        return sx / max_x, sy / max_y

    radio_x, radio_y = get_radio()
    process = subprocess.Popen(
        ['adb', 'shell', 'getevent', '-lt', device],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True
    )
    flag = False
    x, y, t = 0, 0, 0
    x0, y0, t0 = 0, 0, 0
    track_list = []
    try:
        for line in process.stdout:
            if flag_match := re.findall(flag_pattern, line):
                ts, action = flag_match[0]
                if action == 'DOWN':
                    flag = True
                elif action == 'UP':
                    logger.debug(track_list)
                    put2queue(track_list.copy())
                    track_list.clear()
                    x0, y0, t0 = 0, 0, 0
                    x, y, t = 0, 0, 0
                    flag = False
            if not flag:
                continue
            xy_match = re.findall(xy_pattern, line)
            if not xy_match:
                continue
            ts, xy, value_hex = xy_match[0]
            value = int(value_hex, 16)
            if xy == 'X':
                x = value
                t = int(float(ts) * 1000)
            elif xy == 'Y':
                if t != int(float(ts) * 1000):
                    x = 0
                    continue
                y = value
            if all([x, y, t]):
                if not any([x0, y0, t0]):
                    x0, y0, t0 = x, y, t
                track_list.append(((x - x0) * radio_x, (y - y0) * radio_y, t - t0))
                x, y, t = 0, 0, 0
    except KeyboardInterrupt:
        pass
    finally:
        # 确保进程结束
        process.terminate()


def listen_pc(
        interval=0.001
):
    require("pynput==1.7.6")
    from pynput.mouse import Listener
    from pynput.keyboard import Listener as keybord_listener
    # 可变化的原点
    x0, y0, t0 = 0, 0, 0
    # 录制标志
    flag = False
    # 临时队列
    res = []

    def on_move(x, y):
        nonlocal t0
        now = int(time.time() * 1000)
        if t0 == 0:
            t0 = now
        res.append((int(x - x0), int(y - y0), int(time.time() * 1000) - t0))
        time.sleep(interval)

    def on_click(x, y, button, pressed):  # noqa
        nonlocal flag, res, x0, y0, t0
        flag = not flag
        logger.debug('开始录制' if flag else '结束录制')
        if flag:
            x0, y0, t0 = x, y, 0
            res.clear()
        else:
            now = res.copy()
            logger.info(now)
            put2queue(now)

    def on_press(key):
        print(key, Key.esc)
        print(key == Key.esc)
        if key == Key.esc:
            os._exit(0)

    kl = keybord_listener(on_press=on_press)
    kl.start()

    with Listener(on_move=on_move, on_click=on_click) as listen:
        logger.debug(f"可开始录制, 按 esc 强制退出(不保存)")
        listen.join()


def save_data(data, filename=None, save=None, min_length=100, tag='default', save_mode='override', **kwargs):
    data = [_ for _ in data if len(_) > min_length]
    if not data:
        logger.debug(f'未有符合条件的轨迹,不保存')
        return
    if filename:
        logger.debug(f'保存历史轨迹至文件 {filename}: {json.dumps(data)}')
        if save_mode == 'override':
            with open(filename, 'w', encoding='utf-8') as f:
                json.dump(data, f)
        else:
            try:
                with open(filename, 'r', encoding='utf-8') as f:
                    data = json.load(f)
            except Exception:  # noqa
                data = []
            data.extend(data)
            with open(filename, 'w', encoding='utf-8') as f:
                json.dump(data, f)
    elif callable(save):
        logger.debug(f'调用回调函数: {json.dumps(data)}')
        magic.result(save, namespace=dict(track_data=data, interval=kwargs.get('interval'), tag=tag))


def get_track(
        distance: int,
        track_list: list = None,
        filename="",
        max_dis=1000,
        min_node=100,
        tag=None
):
    """
    获取随机轨迹
    :param filename: 轨迹的文件
    :param distance: 需生成的轨迹长度, 需要小于 max_dis 以及 track_list中的最小长度的三分之一, 用以保证轨迹的随机性
    :param track_list: 自带轨迹, 在该轨迹中随机选取一段轨迹
    :param max_dis: 限制最大传入长度
    :param min_node: 最少的节点数, 用以保证轨迹的随机性
    :param tag: 
    :return:
    """

    def get_track_list():
        assert filename, "未传入轨迹或文件"
        with open(filename, 'r') as track_f:
            data = json.load(track_f)
        assert data, f"轨迹文件 {filename} 为空!"
        return data

    def binary_search(lis: list, x, left=None, right=None, ):
        """
        二分查找
        :param lis: 被查找的队列
        :param x: 目标值
        :param left: 起始点
        :param right: 结束点
        :return:
        """
        left = left or 0
        right = right or len(lis)
        while left <= right:
            mid = left + (right - left) // 2
            if lis[mid] == x:
                return mid
            elif lis[mid] > x:
                right = mid - 1
            else:
                left = mid + 1
        return right

    def get_one_track():
        """
        获取一条轨迹
        :return:
        """
        one_track_xy = choice(track_list)
        if len(one_track_xy) < min_node:
            logger.debug(f'轨迹节点数小于 {min_node}, 重新获取轨迹')
            return get_one_track()
        one_track = [_[0] for _ in one_track_xy]
        start_index = randint(0, len(one_track) - 2)
        last = one_track[-1]
        start = one_track[start_index]
        return last, start, one_track, one_track_xy, start_index

    def get_target_track():
        """
        获取目标轨迹
        :return:
        """
        last, start, one_track, one_track_xy, start_index = get_one_track()
        while last - start < distance:
            last, start, one_track, one_track_xy, start_index = get_one_track()
        target = start + distance
        target_index = binary_search(one_track, target)
        _res = one_track_xy[start_index: target_index + 2]
        start_row = _res[0]
        _result = []
        for _ in _res:
            _result.append([_[i] - start_row[i] for i in range(len(_))])
        _distance = target - start_row[0]
        while _result[-2][0] >= _distance:
            _result.pop(-1)
        _result[-1][0] = target - start_row[0]
        return _result

    while not (track_list or local_track[f'$track_list_{tag}']):
        # 只有拿到锁之后才会去获取 track_list 指纹
        if lock.acquire():
            try:
                local_track[f'$track_list_{tag}'] = get_track_list()
            finally:
                lock.release()
        else:
            logger.debug('等待 track_list 获取')
            time.sleep(5)

    track_list = track_list or [magic.json_or_eval(_) for _ in local_track[f'$track_list_{tag}']]

    assert track_list != [], "未获取到符合条件的 track_list !"

    if distance > max_dis:
        raise ValueError('Distance 太大了')

    track = get_target_track()
    return track


if __name__ == '__main__':
    # print(get_track(130, tag='default'))
    # listener(filename='a.json', tag='low1', plat='pc', interval=0.01)
    get_track(filename='a.json', distance=100)



# -*- coding: utf-8 -*-
"""
@File    : cache.py
@Date    : 2025/5/8 下午2:39
@Author  : yintian
@Desc    : 线程安全的缓存管理器
"""
import threading
import time
from itertools import count
from typing import Any, Optional

from ytools import logger
from ytools.utils.date import Arrow


class CacheItem:
    expire_time: Optional[float] = None
    max_uses: Optional[int] = None

    def __init__(self, cache_id, value: Any, expire_time: Optional[float] = None, max_uses: Optional[int] = None, is_empty=False):
        self.is_empty = is_empty
        self.id = cache_id
        self.value = value
        self.set_expire_time(expire_time)
        self.set_max_uses(max_uses)
        self._use_count = count()
        self.use_count = next(self._use_count)
        self.create_time = time.time()

    def __bool__(self):
        return self.is_empty is not None

    def use(self):
        self.use_count = next(self._use_count)

    def check_expired(self) -> bool:
        if self.expire_time is None:
            return True
        return time.time() < self.expire_time

    def check_uses(self):
        if self.max_uses is not None and self.use_count >= self.max_uses:
            return False
        return True

    def can_use(self) -> bool:
        return self.check_expired() and self.check_uses()

    def set_max_uses(self, uses):
        if uses is None:
            return
        if isinstance(uses, int) and uses > 0:
            self.max_uses = uses
        else:
            self.max_uses = None
        logger.debug(f"{self.id} 设置最大使用次数为: {self.max_uses}")

    def set_expire_time(self, expire_time):
        if expire_time is None:
            return
        if isinstance(expire_time, int):
            if expire_time > 10 ** 10:
                self.expire_time = Arrow.get(expire_time).ts(10)
            elif expire_time > 0:
                self.expire_time = Arrow.now().shift(seconds=expire_time).ts(10)
            else:
                self.expire_time = None
        else:
            try:
                self.expire_time = Arrow.get(expire_time).ts(10)
            except Exception as e:
                logger.error(f"{self.id} 设置过期时间 {expire_time} 失败, 错误: {e}")
                self.expire_time = None
        logger.debug(f"{self.id} 设置过期时间为: {self.expire_time}")


class CacheManager(dict):
    def __init__(self, *args, **kwargs):
        self.default_expire_time = kwargs.pop('expire_time', None)
        self.default_max_uses = kwargs.pop('max_uses', None)
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()

    def __setitem__(self, key: str, value: Any) -> None:
        """重写字典的设置方法，设置默认过期时间和使用次数限制"""
        key = str(key)
        with self._lock:
            if isinstance(value, CacheItem):
                super().__setitem__(key, value)
            else:
                super().__setitem__(key, CacheItem(key, value, expire_time=self.default_expire_time, max_uses=self.default_max_uses))

    def __getitem__(self, key: str) -> CacheItem:
        """重写字典的获取方法，增加过期和使用次数检查"""
        with self._lock:
            key = str(key)
            if key not in self:
                return CacheItem(None, None, is_empty=True)

            cache_item: CacheItem = super().__getitem__(key)
            if not cache_item.can_use():
                del self[key]
                return CacheItem(None, None, is_empty=True)

            cache_item.use()
            return cache_item

    def clean(self) -> None:
        """清理所有过期的缓存项"""
        with self._lock:
            expired_keys = [
                key for key, item in self.items()
                if not item.can_use()
            ]
            for key in expired_keys:
                del self[key]

    def set_expire(self, cache_id, expire_time=None, max_uses=None):
        if cache_id not in self:
            return
        expire_time is not None and self.set_expire_time(cache_id, expire_time)
        max_uses is not None and self.set_max_uses(cache_id, max_uses)

    def set_max_uses(self, cache_id, max_uses):
        if cache_id not in self:
            return
        cache_item: CacheItem = super().__getitem__(cache_id)
        cache_item.set_max_uses(max_uses)

    def set_expire_time(self, cache_id, expire_time):
        if cache_id not in self:
            return
        cache_item: CacheItem = super().__getitem__(cache_id)
        cache_item.set_expire_time(expire_time)


if __name__ == '__main__':
    cm = CacheManager()
    cm['1'] = 3
    cm.set_expire('1', 6, 10)
    cm.pop('1', None)
    for i in range(10):
        print(cm['1'].value)
        time.sleep(1)
